﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivQuad.hpp"
# include "SivQuaternion.hpp"

namespace s3d
{
	/// <summary>
	/// 検出した AR マーカーの情報
	/// </summary>
	struct ARMarker
	{
		/// <summary>
		/// カメラ空間での回転
		/// </summary>
		/// <remarks>
		/// カメラキャリブレーションを設定していない場合、利用できません。
		/// </remarks>
		Quaternion rotation3D;

		/// <summary>
		/// フレーム画像内のマーカーの領域
		/// </summary>
		Quad quad;

		/// <summary>
		/// フレーム画像内のマーカーの中心位置
		/// </summary>
		Vec2 screenPos;

		/// <summary>
		/// カメラ空間での座標
		/// </summary>
		/// <remarks>
		/// カメラキャリブレーションを設定していない場合、利用できません。
		/// </remarks>
		Vec3 pos;

		/// <summary>
		/// フレーム画像内のマーカーの回転
		/// </summary>
		double rotation2D;

		/// <summary>
		/// フレーム画像内のマーカー領域の周の長さ
		/// </summary>
		double perimeter;

		/// <summary>
		/// フレーム画像内のマーカー領域の面積
		/// </summary>
		double area;

		/// <summary>
		/// マーカー ID
		/// </summary>
		int id;
	};

	/// <summary>
	/// AR
	/// </summary>
	/// <remarks>
	/// AR マーカー検出機能を提供します。
	/// </remarks>
	namespace AR
	{
		/// <summary>
		/// AR マーカー画像を作成します。
		/// </summary>
		/// <param name="id">
		/// AR マーカーの ID [0, 1023]
		/// </param>
		/// <param name="tileSize">
		/// 画像の縦横の大きさ（ピクセル）
		/// </param>
		/// <returns>
		/// AR マーカー画像、ID が範囲外の場合空の画像
		/// </returns>
		Image CreateMarker(unsigned id, unsigned tileSize);

		/// <summary>
		/// AR マーカー検出を開始します。
		/// </summary>
		/// <returns>
		/// AR マーカー検出の開始に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Start();

		/// <summary>
		/// AR マーカー検出を終了します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void Stop();

		/// <summary>
		/// AR マーカー検出がアクティブかどうかを示します。
		/// </summary>
		/// <returns>
		/// AR マーカー検出がアクティブな場合 true, それ以外の場合は false
		/// </returns>
		bool IsActive();

		/// <summary>
		/// 検出する AR マーカーのサイズを設定します。
		/// </summary>
		/// <param name="min">
		/// 最小サイズ（ピクセル）
		/// </param>
		/// <param name="max">
		/// 最大サイズ（ピクセル）
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetValidSize(double min, double max);

		/// <summary>
		/// カメラキャリブレーションを設定します。
		/// </summary>
		/// <param name="calibFilePath">
		/// カメラキャリブレーションの設定を記述したファイルのパス
		/// </param>
		/// <param name="markerSizeMeter">
		/// 検出に使うマーカーの実測サイズ（m)
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetCalibration(const FilePath& calibFilePath, double markerSizeMeter);

		/// <summary>
		/// 新しい AR マーカー情報があるかを取得します。
		/// </summary>
		/// <returns>
		/// 新しい AR マーカー情報がある場合 true, それ以外の場合は false
		/// </returns>
		bool HasNewFrame();

		/// <summary>
		/// AR マーカー情報を取得します。
		/// </summary>
		/// <param name="markers">
		/// AR マーカー情報のコピー先
		/// </param>
		/// <param name="mirrored">
		/// AR マーカーの座標を反転するかの指定
		/// </param>
		/// <returns>
		/// 取得に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool GetFrame(Array<ARMarker>& markers, bool mirrored = false);
	}
}
