﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivArray.hpp"
# include "SivString.hpp"

namespace s3d
{
	/// <summary>
	/// 圧縮・展開
	/// </summary>
	/// <remarks>
	/// ファイルとフォルダを圧縮・展開する機能を提供します。
	/// </remarks>
	namespace Compression
	{
		/// <summary>
		/// ファイルをアーカイブします。
		/// </summary>
		/// <param name="fileList">
		/// アーカイブするファイルのパス。半角スペースで区切ると複数指定できる
		/// </param>
		/// <param name="archiveName">
		/// アーカイブファイルのパス。拡張子は .zip か .7z
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// アーカイブファイルの作成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Archive(const String& fileList, const FilePath& archiveName, const String& password = L"");

		/// <summary>
		/// ファイルをアーカイブします。
		/// </summary>
		/// <param name="filePaths">
		/// アーカイブするファイルのパスの一覧
		/// </param>
		/// <param name="archiveName">
		/// アーカイブファイルのパス。拡張子は .zip か .7z
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// アーカイブファイルの作成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Archive(const Array<FilePath>& filePaths, const FilePath& archiveName, const String& password = L"");

		/// <summary>
		/// アーカイブファイルにファイルを追加します。
		/// </summary>
		/// <param name="fileList">
		/// 追加するファイルのパス。半角スペースで区切ると複数指定できる
		/// </param>
		/// <param name="archiveName">
		/// アーカイブファイルのパス
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// アーカイブファイルへの追加に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Add(const String& fileList, const FilePath& archiveName, const String& password = L"");

		/// <summary>
		/// アーカイブファイルにファイルを追加します。
		/// </summary>
		/// <param name="filePaths">
		/// アーカイブに追加するファイルのパスの一覧
		/// </param>
		/// <param name="archiveName">
		/// アーカイブファイルのパス
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// アーカイブファイルへの追加に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Add(const Array<FilePath>& filePaths, const FilePath& archiveName, const String& password = L"");

		/// <summary>
		/// アーカイブファイルを展開します。
		/// </summary>
		/// <param name="archiveName">
		/// アーカイブファイルのパス
		/// </param>
		/// <param name="outputDirectory">
		/// 出力先のディレクトリパス
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// 展開に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Extract(const FilePath& archiveName, const FilePath& outputDirectory = L"", const String& password = L"");

		/// <summary>
		/// アーカイブファイルの一部を展開します。
		/// </summary>
		/// <param name="archiveName">
		/// アーカイブファイルのパス
		/// </param>
		/// <param name="outputDirectory">
		/// 出力先のディレクトリパス
		/// </param>
		/// <param name="file">
		/// 展開するファイルのパス
		/// </param>
		/// <param name="password">
		/// パスワード。使用しない場合は空の文字列
		/// </param>
		/// <returns>
		/// 展開に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool ExtractPart(const FilePath& archiveName, const FilePath& outputDirectory, const FilePath& file, const String& password = L"");

		/// <summary>
		/// アーカイブファイルの一覧を取得します。
		/// </summary>
		/// <param name="archiveName">
		/// アーカイブファイルのパス
		/// </param>
		/// <returns>
		/// アーカイブファイルの一覧
		/// </returns>
		Array<FilePath> GetContents(const FilePath& archiveName);

		/// <summary>
		/// アーカイブに使用できるパスワードかどうかを調べます。
		/// </summary>
		/// <param name="password">
		/// パスワード
		/// </param>
		/// <returns>
		/// アーカイブに使用できる場合 true, それ以外の場合は false
		/// </returns>
		bool IsValidPassword(const String& password);

		/// <summary>
		/// アーカイブのコマンド命令を実行します。
		/// </summary>
		/// <param name="command">
		/// コマンド
		/// </param>
		/// <returns>
		/// 実行に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Command(const String& command);
	}
}
