﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivTexture.hpp"

namespace s3d
{
	/// <summary>
	/// 動的テクスチャ
	/// </summary>
	/// <remarks>
	/// 描画可能な画像です。
	/// Texture と異なり、メモリを再確保することなく内容を更新できます。
	/// </remarks>
	class DynamicTexture : public Texture
	{
	private:

		bool
			fill(
			unsigned width,
			unsigned height,
			const void* src,
			unsigned srcSize,
			unsigned stride,
			TextureFormat format,
			bool wait
			);

	public:

		DynamicTexture();

		DynamicTexture(unsigned width, unsigned height, TextureFormat format = TextureFormat::R8G8B8A8_Unorm);

		DynamicTexture(unsigned width, unsigned height, const Color& color);

		DynamicTexture(const Image& image);

		template<typename Type> explicit DynamicTexture(const CustomImage<Type>& image) 
			: Texture(
				image.width,
				image.height,
				image.data(),
				image.dataSize,
				image.stride,
				CustomImage<Type>::Format,
				Texture::_modeDyanmic()
			) {}

		bool fill(const Color& color);

		bool fill(const Color& color, const Rect& rect);

		/// <summary>
		/// 動的テクスチャの中身を同じ大きさの画像で更新します。
		/// 動的テクスチャが空の場合は新しい動的テクスチャを画像から作成します。
		/// </summary>
		/// <param name="image">
		/// 画像
		/// </param>
		/// <returns>
		/// 動的テクスチャの更新か作成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool fill(const Image& image);
		
		template<typename Type> bool fill(const CustomImage<Type>& image)
		{
			return fill(image.width, image.height, image.data(), image.dataSize, image.stride, CustomImage<Type>::Format, true);
		}

		bool fill(const Image& image, const Rect& rect);

		bool tryFill(const Color& color);

		bool tryFill(const Color& color, const Rect& rect);

		/// <summary>
		/// 動的テクスチャの中身を同じ大きさの画像で更新します。
		/// 動的テクスチャが空の場合は新しい動的テクスチャを画像から作成します。
		/// fill() と異なり、テクスチャがビジー状態の時は更新を回避し、パフォーマンスの低下を防ぎます。
		/// </summary>
		/// <param name="image">
		/// 画像
		/// </param>
		/// <returns>
		/// 動的テクスチャの更新か作成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool tryFill(const Image& image);

		template<typename Type> bool tryFill(const CustomImage<Type>& image)
		{
			return fill(image.width, image.height, image.data(), image.dataSize, image.stride, CustomImage<Type>::Format, false);
		}

		bool tryFill(const Image& image, const Rect& rect);
	};
}
