﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <functional>
# include <cmath>
# include "SivMathConstants.hpp"

namespace s3d
{
	/// <summary>
	/// イージング
	/// </summary>
	/// <remarks>
	/// イージング関数を提供します。
	/// </remarks>
	namespace Easing
	{
		inline double Linear(double t)
		{
			return t;
		}

		inline double Sine(double t)
		{
			return 1.0 - std::cos(t * Math::HalfPi);
		}

		inline 	double Quad(double t)
		{
			return t * t;
		}

		inline double Cubic(double t)
		{
			return t * t * t;
		}

		inline double Quart(double t)
		{
			return (t * t) * (t * t);
		}

		inline double Quint(double t)
		{
			return (t * t) * (t * t) * t;
		}

		inline double Expo(double t)
		{
			return t == 0.0 ? 0.0 : std::exp2(10.0*(t - 1.0));
		}

		inline double Circ(double t)
		{
			return 1.0 - std::sqrt(1.0 - t * t);
		}

		inline double Back(double t)
		{
			return t * t * (2.70158 * t - 1.70158);
		}

		inline double ElasticAP(double t, double a, double p)
		{
			if (t == 0)
				return 0.0;
			else if (t == 1)
				return 1.0;

			double s;

			if (a < 1.0)
			{
				a = 1.0;

				s = p / 4.0f;
			}
			else
			{
				s = p / Math::TwoPi * std::asin(1.0 / a);
			}

			t -= 1.0;

			return -(a * std::exp2(10.0 * t) * std::sin((t - s) * Math::TwoPi / p));
		}

		inline double Elastic(double t)
		{
			return ElasticAP(t, 1.0, 0.3);
		}

		inline double Bounce(double t)
		{
			if (t < 1 / 2.75)
			{
				return (7.5625*t*t);
			}
			else if (t < 2 / 2.75)
			{
				t -= 1.5 / 2.75;

				return(7.5625*t*t + 0.75);
			}
			else if (t < 2.5 / 2.75)
			{
				t -= 2.25 / 2.75;

				return (7.5625*t*t + 0.9375);
			}
			else
			{
				t -= 2.625 / 2.75;

				return (7.5625*t*t + 0.984375);
			}
		}

		inline double EaseIn(std::function<double(double)> f, double t)
		{
			return f(t);
		}

		inline double EaseOut(std::function<double(double)> f, double t)
		{
			return 1.0 - f(1.0 - t);
		}

		inline double EaseInOut(std::function<double(double)> f, double t)
		{
			return (t < 0.5) ? f(2.0 * t) * 0.5 : 0.5 + EaseOut(f, 2.0 * t - 1.0) * 0.5;
		}

		template<typename Type> inline Type EaseIn(const Type& start, const Type& end, std::function<double(double)> f, double t)
		{
			return Math::Lerp(start, end, f(t));
		}

		template<typename Type> inline Type EaseOut(const Type& start, const Type& end, std::function<double(double)> f, double t)
		{
			return Math::Lerp(start, end, EaseOut(f, t));
		}

		template<typename Type> inline Type EaseInOut(const Type& start, const Type& end, std::function<double(double)> f, double t)
		{
			return Math::Lerp(start, end, EaseInOut(f, t));
		}
	}

# ifndef NO_EASING_USING
	using Easing::EaseIn;
	using Easing::EaseOut;
	using Easing::EaseInOut;
# endif
}
