﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivString.hpp"
# include "SivArchivedFileReader.hpp"

namespace s3d
{
	/// <summary>
	/// アーカイブ
	/// </summary>
	/// <remarks>
	/// ファイルアーカイブを作成する機能を提供します。
	/// </remarks>
	namespace Archive
	{
		/// <summary>
		/// ファイルアーカイブを作成します。
		/// </summary>
		/// <param name="from">
		/// アーカイブ化するディレクトリ
		/// </param>
		/// <param name="to">
		/// 保存するアーカイブファイル名
		/// </param>
		/// <returns>
		/// ファイルアーカイブの作成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Create(const FilePath& from, const FilePath& to);
	}

	/// <summary>
	/// ファイルアーカイブ
	/// </summary>
	class FileArchive
	{
	private:

		class CFileArchive;

		std::shared_ptr<CFileArchive> pFile;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		FileArchive();

		/// <summary>
		/// ファイルアーカイブを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		explicit FileArchive(const FilePath& path);

		/// <summary>
		/// デストラクタ
		/// </summary>
		~FileArchive();

		/// <summary>
		/// ファイルアーカイブを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <returns>
		/// ファイルアーカイブのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path);

		/// <summary>
		/// ファイルアーカイブがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルアーカイブがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const;

		/// <summary>
		/// アーカイブされたファイル名の一覧を返します。
		/// </summary>
		/// <returns>
		/// アーカイブされたファイル名の一覧
		/// </returns>
		Array<FilePath> contents() const;

		/// <summary>
		/// ファイルアーカイブがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルアーカイブがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// アーカイブされたファイルをロードします。
		/// </summary>
		/// <param name="path">
		/// アーカイブされたファイルのパス
		/// </param>
		/// <returns>
		/// アーカイブされたファイルの ArchivedFileReader
		/// </returns>
		ArchivedFileReader load(const FilePath& path);
	};
}
