﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivBlendState.hpp"
# include "SivRasterizerState.hpp"
# include "SivDepthStencilState.hpp"
# include "SivSamplerState.hpp"
# include "SivRenderState.hpp"

namespace s3d
{
	/// <summary>
	/// 2D グラフィックス
	/// </summary>
	/// <remarks>
	/// 2D　グラフィックスに関連する機能を提供します。
	/// </remarks>
	namespace Graphics2D
	{
		/// <summary>
		/// 2D 描画時の変換行列を設定します。
		/// </summary>
		/// <remarks>
		/// GUI と Println 等による画面表示には適用されません。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetTransform(const Mat3x2& matrix);

		/// <summary>
		/// 現在の 2D 描画時の変換行列を取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画時の変換行列
		/// </returns>
		Mat3x2 GetTransform();

		void SetUVTransform(const Float4& transform);

		Float4 GetUVTransform();

		/// <summary>
		/// 2D 描画のブレンドステートを設定します。
		/// </summary>
		/// <param name="blendState">
		/// ブレンドステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetBlendState(const BlendState& blendState);

		/// <summary>
		/// 現在の 2D 描画のブレンドステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のブレンドステート
		/// </returns>
		BlendState GetBlandState();

		/// <summary>
		/// 2D 描画のラスタライザーステートを設定します。
		/// </summary>
		/// <param name="rasterizerState">
		/// ラスタライザーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRasterizerState(const RasterizerState& rasterizerState);

		/// <summary>
		/// 現在の 2D 描画のラスタライザーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のラスタライザーステート
		/// </returns>
		RasterizerState GetRasterizerState();

		/// <summary>
		/// 2D 描画のステンシルステートを設定します。
		/// </summary>
		/// <param name="stencilState">
		/// ステンシルステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilState(const StencilState& stencilState);

		/// <summary>
		/// 現在の 2D 描画のステンシルステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のステンシルステート
		/// </returns>
		StencilState GetStencilState();

		/// <summary>
		/// 2D 描画のステンシル値を設定します。
		/// </summary>
		/// <param name="value">
		/// ステンシル値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilValue(unsigned char value);

		/// <summary>
		/// 現在の 2D 描画のステンシル値を取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のステンシル値
		/// </returns>
		unsigned char GetStencilValue();

		/// <summary>
		/// 2D 描画のサンプラーステートを設定します。
		/// </summary>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(const SamplerState& samplerState);

		/// <summary>
		/// 2D 描画のサンプラーステートをスロットを指定して設定します。
		/// </summary>
		/// <param name="slot">
		/// スロット [0 または 1]
		/// </param>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(uint32 slot, const SamplerState& samplerState);

		/// <summary>
		/// 現在の 2D 描画のサンプラーステートを取得します。
		/// </summary>
		/// <param name="slot">
		/// スロット [0 または 1]
		/// </param>
		/// <returns>
		/// 現在の 2D 描画のサンプラーステート
		/// </returns>
		SamplerState GetSamplerState(uint32 slot = 0);

		/// <summary>
		/// 現在の 2D 描画のレンダーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のレンダーステート
		/// </returns>
		RenderState GetRenderState();

		/// <summary>
		/// 2D 描画のレンダーステートを設定します。
		/// </summary>
		/// <param name="state">
		/// レンダーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRenderState(const RenderState& state);



		bool BeginShader(const PixelShader& ps);

		void EndShader();

		void SetConstant(ShaderStage stage, uint32 slot, const ConstantBufferBase& buffer, const float* data, uint32 num_vectors);

		template<typename Type>
		inline void SetConstant(ShaderStage stage, uint32 slot, const ConstantBuffer<Type>& buffer)
		{
			SetConstant(stage, slot, buffer.base(), buffer.getPtr(), sizeof(Type) / sizeof(__m128));
		}
	}
}
