﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <cassert>
# include "SivTypes.hpp"
# include "SivArray.hpp"
# include "SivUtility.hpp"
# include "SivPropertyMacro.hpp"

namespace s3d
{
	/// <summary>
	/// 可変長二次元配列
	/// </summary>
	template<class Type>
	class Grid
	{
	private:

		using ContainerType = std::conditional_t<std::is_same<Type, bool>::value, std::basic_string<Type>, Array<Type>>;

		using iterator = typename ContainerType::iterator;

		using const_iterator = typename ContainerType::const_iterator;

		using reverse_iterator = typename ContainerType::reverse_iterator;

		using const_reverse_iterator = typename ContainerType::const_reverse_iterator;

		using pointer = typename ContainerType::pointer;

		using const_pointer = typename ContainerType::const_pointer;

		ContainerType m_data;

		int32 m_width = 0;

		int32 m_height = 0;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Grid() = default;

		/// <summary>
		/// 幅と高さを指定して二次元配列を作成します。
		/// </summary>
		/// <param name="w">
		/// 幅(列数)
		/// </param>
		/// <param name="h">
		/// 高さ(行数)
		/// </param>
		Grid(int32 w, int32 h)
			: m_data(w * h, Type())
			, m_width{ w }
			, m_height{ h } {}

		/// <summary>
		/// 幅と高さと初期値を指定して二次元配列を作成します。
		/// </summary>
		/// <param name="w">
		/// 幅(列数)
		/// </param>
		/// <param name="h">
		/// 高さ(行数)
		/// </param>
		/// <param name="val">
		/// 初期値
		/// </param>
		Grid(int32 w, int32 h, const Type& val)
			: m_data(w*h, val)
			, m_width{ w }
			, m_height{ h } {}

		/// <summary>
		/// コピーコンストラクタ
		/// </summary>
		Grid(const Grid& g) = default;

		/// <summary>
		/// ムーブコンストラクタ
		/// </summary>
		Grid(Grid&& g)
		{
			m_width = g.m_width;

			m_height = g.m_height;

			m_data = std::move(g.m_data);

			g.m_width = g.m_height = 0;
		}

		/// <summary>
		/// 指定した位置の要素への参照を返します。
		/// </summary>
		/// <param name="y">
		/// 位置(行)
		/// </param>
		/// <param name="x">
		/// 位置(列)
		/// </param>
		/// <exception cref="std::out_of_range">
		/// 範囲外アクセスの場合 throw されます。
		/// </exception>
		/// <returns>
		/// 指定した位置の要素への参照
		/// </returns>
		Type& at(int32 y, int32 x)
		{
			if (!inBounds(y, x))
			{
				throw std::out_of_range("Grid::at");
			}

			return m_data[y*m_width + x];
		}

		/// <summary>
		/// 指定した位置の要素への参照を返します。
		/// </summary>
		/// <param name="y">
		/// 位置(行)
		/// </param>
		/// <param name="x">
		/// 位置(列)
		/// </param>
		/// <exception cref="std::out_of_range">
		/// 範囲外アクセスの場合 throw されます。
		/// </exception>
		/// <returns>
		/// 指定した位置の要素への参照
		/// </returns>
		const Type& at(int32 y, int32 x) const
		{
			if (!inBounds(y, x))
			{
				throw std::out_of_range("Grid::at");
			}

			return m_data[y*m_width + x];
		}

		/// <summary>
		/// 指定した行の先頭ポインタを返します。
		/// </summary>
		/// <param name="index">
		/// 位置(行)
		/// </param>
		/// <returns>
		/// grid[y][x] で指定した要素にアクセスします。
		/// </returns>
		/// <returns>
		/// 指定した行の先頭ポインタ
		/// </returns>
		Type* operator[] (size_t index) { return &m_data[index*m_width]; }

		/// <summary>
		/// 指定した行の先頭ポインタを返します。
		/// </summary>
		/// <param name="index">
		/// 位置(行)
		/// </param>
		/// <returns>
		/// grid[y][x] で指定した要素にアクセスします。
		/// </returns>
		/// <returns>
		/// 指定した行の先頭ポインタ
		/// </returns>
		const Type* operator[] (size_t index) const { return &m_data[index*m_width]; }

		/// <summary>
		/// 指定した位置が二次元配列の範囲内かを返します。
		/// </summary>
		/// <param name="y">
		/// 位置(行)
		/// </param>
		/// <param name="x">
		/// 位置(列)
		/// </param>
		/// <returns>
		/// 二次元配列の範囲内である場合 true, それ以外の場合は false
		/// </returns>
		bool inBounds(int32 y, int32 x) const
		{
			return InRange(y, 0, m_height) && InRange(x, 0, m_width);
		}

		/// <summary>
		/// 二次元配列が空であるかを示します。
		/// </summary>
		Property_Get(bool, isEmpty) const { return m_data.empty(); }

		/// <summary>
		/// 二次元配列の幅(列数)を示します。
		/// </summary>
		Property_Get(int32, width) const { return m_width; }

		/// <summary>
		/// 二次元配列の高さ(行数)を示します。
		/// </summary>
		Property_Get(int32, height) const { return m_height; }

		/// <summary>
		/// 二次元配列のメモリサイズをバイト単位で示します。
		/// </summary>
		Property_Get(size_t, memorySize) const { return m_data.size() * sizeof(Type); }

		/// <summary>
		/// 二次元配列の先頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭位置のイテレータ
		/// </returns>
		iterator begin() { return m_data.begin(); }

		/// <summary>
		/// 二次元配列の終了頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の終了位置のイテレータ
		/// </returns>
		iterator end() { return m_data.end(); }

		/// <summary>
		/// 二次元配列の先頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭位置のイテレータ
		/// </returns>
		const_iterator begin() const { return m_data.cbegin(); }

		/// <summary>
		/// 二次元配列の終了頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の終了位置のイテレータ
		/// </returns>
		const_iterator end() const { return m_data.cend(); }

		/// <summary>
		/// 二次元配列の末尾のリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の末尾のリバースイテレータ
		/// </returns>
		reverse_iterator rbegin() { return m_data.rbegin(); }

		/// <summary>
		/// 二次元配列の先頭の前へのリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭の前へのリバースイテレータ
		/// </returns>
		reverse_iterator rend() { return m_data.rend(); }

		/// <summary>
		/// 二次元配列の末尾のリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の末尾のリバースイテレータ
		/// </returns>
		const_reverse_iterator rbegin() const { return m_data.rbegin(); }

		/// <summary>
		/// 二次元配列の先頭の前へのリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭の前へのリバースイテレータ
		/// </returns>
		const_reverse_iterator rend() const { return m_data.rend(); }

		/// <summary>
		/// 二次元配列の先頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭位置のイテレータ
		/// </returns>
		const_iterator cbegin() const { return m_data.cbegin(); }

		/// <summary>
		/// 二次元配列の終了頭位置のイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の終了位置のイテレータ
		/// </returns>
		const_iterator cend() const { return m_data.cend(); }

		/// <summary>
		/// 二次元配列の末尾のリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の末尾のリバースイテレータ
		/// </returns>
		const_reverse_iterator crbegin() const { return m_data.crbegin(); }

		/// <summary>
		/// 二次元配列の先頭の前へのリバースイテレータを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭の前へのリバースイテレータ
		/// </returns>
		const_reverse_iterator crend() const { return m_data.crend(); }

		/// <summary>
		/// 二次元配列の不要なメモリを解放します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void shrink_to_fit() { m_data.shrink_to_fit(); }

		/// <summary>
		/// 要素をすべて消去し、メモリを解放します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void release() { clear(); shrink_to_fit(); }

		/// <summary>
		/// 末尾に行を追加します。
		/// </summary>
		/// <param name="value">
		/// 追加される行の要素
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void push_row(const Type& value)
		{
			++m_height;

			m_data.resize(m_data.size() + m_width, value);
		}

		/// <summary>
		/// 末尾の行を削除します。
		/// </summary>
		/// <exception cref="std::out_of_range">
		/// 削除する行が無い場合 throw されます。
		/// </exception>
		/// <returns>
		/// なし
		/// </returns>
		void pop_row()
		{
			if (m_height == 0)
			{
				throw std::out_of_range("Grid::pop_row");
			}

			--m_height;

			m_data.resize(m_data.size() - m_width);
		}

		/// <summary>
		/// 二次元配列用のメモリを指定したサイズで確保します。
		/// </summary>
		/// <param name="w">
		/// 確保する幅(列数)
		/// </param>
		/// <param name="h">
		/// 確保する高さ(行数)
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void reserve(int32 w, int32 h) { m_data.reserve(w*h); }

		/// <summary>
		/// 二次元配列のサイズを変更します。
		/// </summary>
		/// <param name="w">
		/// 新しい幅(列数)
		/// </param>
		/// <param name="h">
		/// 新しい高さ(行数)
		/// </param>
		/// <remarks>
		/// サイズが変更された場合、すべての要素の値が不定になります。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void resize(int32 w, int32 h) { resize(w, h, Type()); }

		/// <summary>
		/// 二次元配列のサイズを変更します。
		/// </summary>
		/// <param name="w">
		/// 新しい幅(列数)
		/// </param>
		/// <param name="h">
		/// 新しい高さ(行数)
		/// </param>
		/// <param name="val">
		/// 要素を初期化する値
		/// </param>
		/// <remarks>
		/// サイズが変更された場合、要素はすべて指定した値で初期化されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void resize(int32 w, int32 h, const Type& val)
		{
			if (w == m_width && h == m_height)
			{
				return;
			}

			m_width = w;

			m_height = h;

			m_data.assign(w * h, val);
		}

		/// <summary>
		/// 要素を消去し、空の二次元配列にします。
		/// </summary>
		/// <remarks>
		/// メモリを解放したい場合は shrink_to_fit() を呼びます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void clear()
		{
			m_data.clear();

			m_width = m_height = 0;
		}

		/// <summary>
		/// 要素の個数を返します。
		/// </summary>
		/// <returns>
		/// 要素の個数
		/// </returns>
		size_t size() const { return m_data.size(); }

		/// <summary>
		/// 二次元配列の先頭のポインタを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭へのポインタ
		/// </returns>
		pointer data() { return &m_data[0]; }

		/// <summary>
		/// 二次元配列の先頭のポインタを返します。
		/// </summary>
		/// <returns>
		/// 二次元配列の先頭へのポインタ
		/// </returns>
		const_pointer data() const { return m_data.data(); }

		/// <summary>
		/// Array を返します。
		/// </summary>
		/// <returns>
		/// Array
		/// </returns>
		const ContainerType& getArray() const { return m_data; }

		/// <summary>
		/// すべての要素に指定した値を代入します。
		/// </summary>
		/// <param name="val">
		/// 代入する値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void fill(const Type& val) { std::fill(m_data.begin(), m_data.end(), val); }
	};
}
