﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivParse.hpp"
# include "SivOptional.hpp"
# include "SivString.hpp"
# include "SivFormat.hpp"
# include "SivPropertyMacro.hpp"
# include "SivJSONValue.hpp"

namespace s3d
{
	/// <summary>
	/// JSON ファイルの読み込み
	/// </summary>
	class JSONReader
	{
	private:

		class CJSONReader;

		std::shared_ptr<CJSONReader> pReader;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		JSONReader();

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		explicit JSONReader(const FilePath& path);

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="reader">
		/// IReader
		/// </param>
		template <typename Reader, typename = std::enable_if_t<std::is_base_of<IReader, Reader>::value>> explicit JSONReader(Reader&& reader)
			: JSONReader()
		{
			open(std::move(reader));
		}

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="reader">
		/// IReader
		/// </param>
		explicit JSONReader(const std::shared_ptr<IReader>& reader);

		/// <summary>
		/// デストラクタ
		/// </summary>
		~JSONReader();

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path);

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="reader">
		/// IReader
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		template <typename Reader, typename = std::enable_if_t<std::is_base_of<IReader, Reader>::value>> bool open(Reader&& reader)
		{
			return open(std::make_shared<Reader>(std::move(reader)));
		}

		/// <summary>
		/// JSON ファイルを開きます。
		/// </summary>
		/// <param name="reader">
		/// IReader
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const std::shared_ptr<IReader>& reader);

		/// <summary>
		/// JSON ファイルをクローズします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void close();

		/// <summary>
		/// JSON ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const;

		/// <summary>
		/// JSON ファイルの内容が変更されたかを返します。
		/// </summary>
		/// <remarks>
		/// ファイルの内容が変更された場合、reload() を呼ぶまで true を返します。
		/// </remarks>
		/// <returns>
		/// CSV ファイルの内容が変更された場合 true, それ以外の場合は false
		/// </returns>
		bool hasChanged() const;

		/// <summary>
		/// JSON ファイルが変更された場合、データの内容を更新します。
		/// </summary>
		/// <returns>
		/// 更新に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool reload();

		/// <summary>
		/// JSON ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// ルートの JSON データを返します。
		/// </summary>
		/// <returns>
		/// ルートの JSON データ
		/// </returns>
		const JSONValue& root() const;

		/// <summary>
		/// JSON データを取得します。
		/// </summary>
		/// <param name="str">
		/// キー（L"KEYA" や L"KEYA.KEYB.KEYC" といった形式）
		/// </param>
		/// <returns>
		/// JSON データ
		/// </returns>
		const JSONValue& operator[](const String& str) const;

		/// <summary>
		/// JSON ファイルのパスを示します。
		/// </summary>
		Property_Get(FilePath, path) const;
	};
}


namespace std
{
	template<> inline void swap(s3d::JSONValue& x, s3d::JSONValue& y)
	{
		x.swap(y);
	}
}
