﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivPropertyMacro.hpp"
# include "SivArray.hpp"
# include "SivUtility.hpp"

namespace s3d
{
	enum class KeyDeviceType : uint8
	{
		Keyboard,

		Gamepad,

		XInput,
	};

	/// <summary>
	/// キー
	/// </summary>
	/// <remarks>
	/// 入力デバイスのキーやボタンを示します。
	/// </remarks>
	class Key
	{
	private:

		KeyDeviceType m_deviceType = KeyDeviceType::Keyboard;

		uint8 m_userIndex = 0;

		uint8 m_code = 0;

		uint8 m_reserved = 0;

	public:

		struct VirtualKeyCode{};

		struct GamepadKeyCode{};

		struct XInputKeyCode{};

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Key() = default;

		/// <summary>
		/// アルファベットか数字でキーを指定します。
		/// </summary>
		/// <param name="alnum">
		/// アルファベットか数字
		/// </param>
		explicit Key(wchar alnum);

		/// <summary>
		/// 仮想キーコードでキーを指定します。
		/// </summary>
		/// <param name="code">
		/// 仮想キーコード
		/// </param>
		Key(uint8 code, const VirtualKeyCode&)
			: m_deviceType(KeyDeviceType::Keyboard)
			, m_code(code) {}

		Key(uint32 userIndex, uint8 code, const GamepadKeyCode&)
			: m_deviceType(KeyDeviceType::Gamepad)
			, m_userIndex(static_cast<uint8>(userIndex))
			, m_code(code) {}

		Key(uint32 userIndex, uint8 code, const XInputKeyCode&)
			: m_deviceType(KeyDeviceType::XInput)
			, m_userIndex(static_cast<uint8>(userIndex))
			, m_code(code) {}

		/// <summary>
		/// キーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const;

		/// <summary>
		/// キーが押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const;

		/// <summary>
		/// キーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const;

		/// <summary>
		/// キーが押されている場合、押し始めからの経過時間を示します。
		/// </summary>
		Property_Get(int32, pressedDuration) const;

		/// <summary>
		/// キーが指すデバイスを示します。
		/// </summary>
		Property_Get(KeyDeviceType, device) const
		{
			return m_deviceType;
		}

		/// <summary>
		/// 仮想キーコードを示します。
		/// </summary>
		Property_Get(uint8, code) const
		{
			return m_code;
		}
	};

	/// <summary>
	/// 2 つのキーが同じキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが同じキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator == (const Key& key1, const Key& key2)
	{
		return key1.device == key2.device && key1.code == key2.code;
	}

	/// <summary>
	/// 2 つのキーが異なるキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが異なるキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator != (const Key& key1, const Key& key2)
	{
		return key1.device != key2.device || key1.code != key2.code;
	}

	/// <summary>
	/// キーの組み合わせ（プラス）
	/// </summary>
	/// <remarks>
	/// 1 つ目のキーが押されている状態のとき 2 つ目のキー入力を調べます。
	/// </remarks>
	class PlusCombinedKeys
	{
	private:

		Key m_key1, m_key2;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		PlusCombinedKeys() = default;

		/// <summary>
		/// キーの組み合わせを設定します。
		/// </summary>
		/// <param name="key1">
		/// 1 つ目のキー
		/// </param>
		/// <param name="key2">
		/// 2 つ目のキー
		/// </param>
		PlusCombinedKeys(Key key1, Key key2)
			: m_key1(key1)
			, m_key2(key2) {}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const
		{
			return m_key1.pressed && m_key2.clicked;
		}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーが押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const
		{
			const auto t1 = m_key1.pressedDuration, t2 = m_key2.pressedDuration;

			if (t1 < t2)
			{
				return false;
			}

			return m_key1.pressed && m_key2.pressed;
		}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const
		{
			return m_key1.pressed && m_key2.released;
		}

		Property_Get(int32, pressedDuration) const
		{
			const auto t1 = m_key1.pressedDuration, t2 = m_key2.pressedDuration;
			const bool pr = m_key1.pressed && m_key2.pressed;

			if (!pr || t1 < t2)
			{
				return 0;
			}

			return t2;
		}

		Property_Get(Key, key1) const
		{
			return m_key1;
		}

		Property_Get(Key, key2) const
		{
			return m_key2;
		}
	};

	/// <summary>
	/// キーの組み合わせ（プラス）を設定します。
	/// </summary>
	/// <param name="key1">
	/// 1 つ目のキー
	/// </param>
	/// <param name="key2">
	/// 2 つ目のキー
	/// </param>
	/// <returns>
	/// キーの組み合わせ（プラス）
	/// </returns>
	inline PlusCombinedKeys operator + (Key key1, Key key2)
	{
		return{ key1, key2 };
	}

	/// <summary>
	/// 2 つのキーが同じキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが同じキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator == (const PlusCombinedKeys& key1, const PlusCombinedKeys& key2)
	{
		return key1.key1 == key2.key1 && key1.key2 == key2.key2;
	}

	/// <summary>
	/// 2 つのキーが異なるキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが異なるキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator != (const PlusCombinedKeys& key1, const PlusCombinedKeys& key2)
	{
		return key1.key1 != key2.key1 || key1.key2 != key2.key2;
	}

	class KeyCombination
	{
	private:

		Array<Key> m_keys;

		Array<PlusCombinedKeys> m_pkeys;

		template <typename Type, typename ... Args>
		void append(const Type& key, const Args& ... args)
		{
			append(key);
			append(args...);
		}

		void append();

		void append(Key key);

		void append(PlusCombinedKeys key);

		void append(const KeyCombination& vkey);

	public:

		template <typename ... Args>
		KeyCombination(const Args& ... args)
		{
			append(args...);
		}

		template <typename Type>
		KeyCombination& operator |= (const Type& key)
		{
			append(key);
			return *this;
		}

		/// <summary>
		/// キーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const;

		/// <summary>
		/// キーが押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const;

		/// <summary>
		/// キーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const;

		Property_Get(int32, num_pressed) const;

		Property_Get(int32, pressedDuration) const;
	};

	inline KeyCombination operator | (Key key1, Key key2)
	{
		return KeyCombination(key1, key2);
	}

	inline KeyCombination operator | (const KeyCombination& key1, Key key2)
	{
		return KeyCombination(key1, key2);
	}

	inline KeyCombination operator | (Key key1, const KeyCombination& key2)
	{
		return KeyCombination(key1, key2);
	}

	inline KeyCombination operator | (const KeyCombination& key1, const KeyCombination& key2)
	{
		return KeyCombination(key1, key2);
	}
}
