﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivArray.hpp"
# include "SivUtility.hpp"
# include "SivPointVector.hpp"
# include "SivLine.hpp"

namespace s3d
{
	/// <summary>
	/// 連続した複数の線分
	/// </summary>
	class LineString
	{
	private:

		Array<Vec2> m_data;

	public:

		LineString() = default;

		explicit LineString(const String& pts)
		{
			std::wistringstream(pts.str()) >> *this;
		}

		explicit LineString(const Array<Point>& pts)
		{
			m_data.reserve(pts.size());

			for (const auto& p : pts)
			{
				m_data.push_back(p);
			}
		}

		explicit LineString(const Array<Vec2>& pts)
			: m_data(pts) {}

		explicit LineString(Array<Vec2>&& pts)
			: m_data(std::move(pts)) {}

		LineString(const Vec2* pts, size_t size)
			: m_data(pts, pts+size) {}

		template<typename Iterator>
		LineString(Iterator first, Iterator last)
			: m_data(first, last) {}

		explicit LineString(std::initializer_list<Vec2> il)
			: LineString(il.begin(), il.size()) {}

		LineString movedBy(double x, double y) const { return movedBy({ x, y }); }

		LineString movedBy(const Vec2& v) const;

		LineString& moveBy(double x, double y) { return moveBy({ x, y }); }

		LineString& moveBy(const Vec2& v);

		Property_Get(bool, isEmpty) const { return m_data.empty(); }

		Property_Get(size_t, num_points) const { return m_data.size(); }

		Property_Get(size_t, num_lines) const { return isEmpty ? 0 : m_data.size() - 1; }

		size_t size() const { return m_data.size(); }

		Vec2& operator [] (size_t index) { return m_data[index]; }

		const Vec2& operator [] (size_t index) const { return m_data[index]; }

		Array<Vec2>::iterator begin() { return m_data.begin(); }

		Array<Vec2>::iterator end() { return m_data.end(); }

		Array<Vec2>::const_iterator begin() const { return m_data.begin(); }

		Array<Vec2>::const_iterator end() const { return m_data.end(); }

		Array<Vec2>& data() { return m_data; }

		const Array<Vec2>& data() const { return m_data; }

		void push_back(const Vec2& point) { m_data.push_back(point); }

		void pop_back() { m_data.pop_back(); }

		void clear() { m_data.clear(); }

		void reserve(size_t size) { m_data.reserve(size); }

		void shrink_to_fit() { m_data.shrink_to_fit(); }

		Line line(uint32 index) const
		{
			return{ m_data[index], m_data[index + 1] };
		}

		LineString substr(size_t offset) const
		{
			return LineString(begin() + offset, end());
		}

		LineString substr(size_t offset, size_t count) const
		{
			return LineString(begin() + offset, begin() + offset + count);
		}

		LineString toCatmullRomSpline(bool isClosedCurve = false, int32 interpolation = 24) const;

		template<typename Shape> bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		void write(Image& image, const Color& color = Palette::Black, bool isClosed = false) const;

		void write(Image& image, int32 thickness, const Color& color = Palette::Black, bool isClosed = false) const;

		void overwrite(Image& image, const Color& color = Palette::Black, bool isClosed = false) const;

		void overwrite(Image& image, int32 thickness, const Color& color = Palette::Black, bool isClosed = false) const;

		void writeCatmullRomSpline(Image& image, int32 thickness = 1, const Color& color = Palette::Black, bool isClosed = false, int32 interpolation = 24) const;

		void overwriteCatmullRomSpline(Image& image, int32 thickness = 1, const Color& color = Palette::Black, bool isClosed = false, int32 interpolation = 24) const;

		void draw(const Color& color = Palette::White, bool isClosedCurve = false) const;

		void draw(double thickness, const Color& color = Palette::White, bool isClosedCurve = false) const;

		void drawCatmullRomSpline(double thickness = 1.0, const Color& color = Palette::White, bool isClosedCurve = false, int interpolation = 24) const;

		Polygon asPolygon(double thickness = 1.0, bool isClosed = false) const;
	};

	template<typename CharType>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const LineString& lineStr
		)
	{
		os << CharType('(');

		bool b = false;

		for (const auto& point : lineStr)
		{
			if (Exchange(b, true))
			{
				os << CharType(',');
			}

			os << point;
		}

		return os << CharType(')');
	}

	template<typename CharType>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		LineString& lineStr
		)
	{
		lineStr.clear();

		CharType unused;

		is >> unused;

		for (;;)
		{
			Vec2 v;

			is >> v;

			if (!is)
			{
				break;
			}

			is >> unused;

			lineStr.push_back(v);

			if (unused == L')')
			{
				break;
			}
		}

		return is;
	}
}
