﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivPointVector.hpp"

namespace s3d
{
	/// <summary>
	/// 3x2 行列
	/// </summary>
	struct Mat3x2
	{
		float _11, _12;
		float _21, _22;
		float _31, _32;

		Mat3x2() = default;

		Mat3x2(float _11, float _12, float _21, float _22, float _31, float _32)
			: _11{ _11 }
			, _12{ _12 }
			, _21{ _21 }
			, _22{ _22 }
			, _31{ _31 }
			, _32{ _32 } {}

		static Mat3x2 Identity()
		{
			Mat3x2 m;
			m._11 = 1.0f;
			m._12 = 0.0f;
			m._21 = 0.0f;
			m._22 = 1.0f;
			m._31 = 0.0f;
			m._32 = 0.0f;
			return m;
		}

		static Mat3x2 Translate(const Float2& v)
		{
			Mat3x2 m;
			m._11 = 1.0;
			m._12 = 0.0;
			m._21 = 0.0;
			m._22 = 1.0;
			m._31 = v.x;
			m._32 = v.y;
			return m;
		}

		static Mat3x2 Translate(double x, double y)
		{
			return Translate({ x, y });
		}

		static Mat3x2 Scale(double s, const Float2& center = Float2::Zero)
		{
			return Scale({ s, s }, center);
		}

		static Mat3x2 Scale(const Float2& scale, const Float2& center = Float2::Zero)
		{
			Mat3x2 m;
			m._11 = scale.x;
			m._12 = 0.0f;
			m._21 = 0.0f;
			m._22 = scale.y;
			m._31 = center.x - scale.x * center.x;
			m._32 = center.y - scale.x * center.y;
			return m;
		}

		static Mat3x2 Scale(double sx, double sy, const Float2& center = Float2::Zero)
		{
			return Scale({ sx, sy }, center);
		}

		static Mat3x2 Rotate(double angle, const Float2& center = Float2::Zero)
		{
			Mat3x2 m;
			const float s = std::sinf(static_cast<float>(angle));
			const float c = std::cosf(static_cast<float>(angle));
			m._11 = c;
			m._12 = s;
			m._21 = -s;
			m._22 = c;
			m._31 = center.x - center.x * c + center.y * s;
			m._32 = center.y - center.x * s - center.y * c;
			return m;
		}

		Mat3x2 translate(const Float2& v) const
		{
			Mat3x2 result;
			result.setProduct(*this, Translate(v));
			return result;
		}

		Mat3x2 translate(double x, double y) const
		{
			return translate({ x, y });
		}

		Mat3x2 scale(double s, const Float2& center = Float2::Zero) const
		{
			return scale({ s, s }, center);
		}

		Mat3x2 scale(const Float2& scale, const Float2& center = Float2::Zero) const
		{
			Mat3x2 result;
			result.setProduct(*this, Scale(scale, center));
			return result;
		}

		Mat3x2 scale(double sx, double sy, const Float2& center = Float2::Zero) const
		{
			return scale({ sx, sy }, center);
		}

		Mat3x2 rotate(double angle, const Float2& center = Float2::Zero) const
		{
			Mat3x2 result;
			result.setProduct(*this, Rotate(angle, center));
			return result;
		}

		float determinant() const
		{
			return (_11 * _22) - (_12 * _21);
		}

		Mat3x2 inverse() const
		{
			const float det = determinant();
			assert(det != 0.0f);
			const float detInv = 1.0f / det;

			Mat3x2 out;
			out._11 = (_22) * detInv;
			out._12 = (-_12) * detInv;
			out._21 = (-_21) * detInv;
			out._22 = (_11) * detInv;
			out._31 = (_21*_32 - _22*_31) * detInv;
			out._32 = (_12*_31 - _11*_32) * detInv;
			return out;
		}

		void setProduct(const Mat3x2 &a, const Mat3x2 &b)
		{
			_11 = a._11 * b._11 + a._12 * b._21;
			_12 = a._11 * b._12 + a._12 * b._22;
			_21 = a._21 * b._11 + a._22 * b._21;
			_22 = a._21 * b._12 + a._22 * b._22;
			_31 = a._31 * b._11 + a._32 * b._21 + b._31;
			_32 = a._31 * b._12 + a._32 * b._22 + b._32;
		}

		Mat3x2 operator*(const Mat3x2 &matrix) const
		{
			Mat3x2 result;
			result.setProduct(*this, matrix);
			return result;
		}

		Float2 transform(const Point& pos) const
		{
			return
			{
				pos.x * _11 + pos.y * _21 + _31,
				pos.x * _12 + pos.y * _22 + _32
			};
		}

		Float2 transform(const Float2& pos) const
		{
			return
			{
				pos.x * _11 + pos.y * _21 + _31,
				pos.x * _12 + pos.y * _22 + _32
			};
		}

		Vec2 transform(const Vec2& pos) const
		{
			return
			{
				pos.x * _11 + pos.y * _21 + _31,
				pos.x * _12 + pos.y * _22 + _32
			};
		}
	};

	template<typename CharType>
	inline
		std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Mat3x2& mat
		)
	{
		return os << CharType('(')
			<< Float2(mat._11, mat._12) << CharType(',')
			<< Float2(mat._21, mat._22) << CharType(',')
			<< Float2(mat._31, mat._32) << CharType(')');
	}

	template<typename CharType>
	inline
		std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Mat3x2& mat
		)
	{
		CharType unused;
		Float2 r1, r2, r3;

		is >> unused
			>> r1 >> unused
			>> r2 >> unused
			>> r3 >> unused;

		mat._11 = r1.x;
		mat._12 = r1.y;
		mat._21 = r2.x;
		mat._22 = r2.y;
		mat._31 = r3.x;
		mat._32 = r3.y;

		return is;
	}
}
