﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivDirectXMathUtility.hpp"
# include "SivQuaternion.hpp"

namespace s3d
{
	/// <summary>
	/// 4x4 行列
	/// </summary>
	struct Mat4x4 : Matrix
	{
		Mat4x4() = default;

		Mat4x4(const Matrix& mat)
			: Matrix(mat){}

		static Mat4x4 __vectorcall Identity()
		{
			return DirectX::XMMatrixIdentity();
		}

		static Mat4x4 __vectorcall Translate(double x, double y, double z)
		{
			return DirectX::XMMatrixTranslation(static_cast<float>(x), static_cast<float>(y), static_cast<float>(z));
		}

		static Mat4x4 __vectorcall Translate(const Vec3& pos)
		{
			return DirectX::XMMatrixTranslation(static_cast<float>(pos.x), static_cast<float>(pos.y), static_cast<float>(pos.z));
		}


		static Mat4x4 __vectorcall Scale(double scale)
		{
			return DirectX::XMMatrixScaling(static_cast<float>(scale), static_cast<float>(scale), static_cast<float>(scale));
		}

		static Mat4x4 __vectorcall Scale(double xs, double ys, double zs)
		{
			return DirectX::XMMatrixScaling(static_cast<float>(xs), static_cast<float>(ys), static_cast<float>(zs));
		}

		static Mat4x4 __vectorcall Scale(const Vec3& scale)
		{
			return DirectX::XMMatrixScaling(static_cast<float>(scale.x), static_cast<float>(scale.y), static_cast<float>(scale.z));
		}


		/**
		*	値を指定したフィールド幅をもつ文字列に変換します。
		*	@param value 変換するデータ
		*	@param padding フィールド幅と詰め文字のペア
		*	@return フィールド幅に従って詰め文字を追加した文字列
		*/
		static Mat4x4 __vectorcall Rotate(double pitch, double yaw, double roll)
		{
			return DirectX::XMMatrixRotationRollPitchYaw(static_cast<float>(pitch), static_cast<float>(yaw), static_cast<float>(roll));
		}

		static Mat4x4 __vectorcall Rotate(const Vec3& pitchYawRoll)
		{
			return DirectX::XMMatrixRotationRollPitchYaw(static_cast<float>(pitchYawRoll.x), static_cast<float>(pitchYawRoll.y), static_cast<float>(pitchYawRoll.z));
		}

		static Mat4x4 __vectorcall Rotate(const Quaternion& quaternion)
		{
			return quaternion.toMatrix();
		}

		static Mat4x4 __vectorcall AffineTransform(const Vec3& scaling, const Vec3& rotation, const Vec3& translation)
		{
			const DirectX::XMVECTOR scale = ToVector(scaling, 1.0f);
			const DirectX::XMVECTOR rotationOrigin = DirectX::g_XMZero;
			const DirectX::XMVECTOR rotate = DirectX::XMQuaternionRotationRollPitchYaw(
				static_cast<float>(rotation.x),
				static_cast<float>(rotation.y),
				static_cast<float>(rotation.z));
			const DirectX::XMVECTOR translate = ToVector(translation, 0.0f);

			return DirectX::XMMatrixAffineTransformation(scale, rotationOrigin, rotate, translate);
		}

		static Mat4x4 __vectorcall AffineTransform(const Vec3& scaling, const Quaternion& rotation, const Vec3& translation)
		{
			const DirectX::XMVECTOR scale = ToVector(scaling, 1.0f);
			const DirectX::XMVECTOR rotationOrigin = DirectX::g_XMZero;
			const DirectX::XMVECTOR translate = ToVector(translation, 0.0f);

			return DirectX::XMMatrixAffineTransformation(scale, rotationOrigin, rotation.component.v, translate);
		}

		static Mat4x4 __vectorcall AffineTransform(double scaling, const Quaternion& rotation, const Vec3& translation)
		{
			const float sf = static_cast<float>(scaling);
			const DirectX::XMVECTOR scale = DirectX::XMVectorSet(sf, sf, sf, 0.0f);
			const DirectX::XMVECTOR rotationOrigin = DirectX::g_XMZero;
			const DirectX::XMVECTOR translate = ToVector(translation, 0.0f);

			return DirectX::XMMatrixAffineTransformation(scale, rotationOrigin, rotation.component.v, translate);
		}

		Mat4x4 __vectorcall translated(double x, double y, double z) const
		{
			return *this * DirectX::XMMatrixTranslation(static_cast<float>(x), static_cast<float>(y), static_cast<float>(z));
		}

		Mat4x4 __vectorcall translated(const Vec3& pos) const
		{
			return *this * DirectX::XMMatrixTranslation(static_cast<float>(pos.x), static_cast<float>(pos.y), static_cast<float>(pos.z));
		}


		Mat4x4 __vectorcall scaled(double scale) const
		{
			return *this * DirectX::XMMatrixScaling(static_cast<float>(scale), static_cast<float>(scale), static_cast<float>(scale));
		}

		Mat4x4 __vectorcall scaled(double xs, double ys, double zs) const
		{
			return *this * DirectX::XMMatrixScaling(static_cast<float>(xs), static_cast<float>(ys), static_cast<float>(zs));
		}

		Mat4x4 __vectorcall scaled(const Vec3& scale) const
		{
			return *this * DirectX::XMMatrixScaling(static_cast<float>(scale.x), static_cast<float>(scale.y), static_cast<float>(scale.z));
		}


		Mat4x4 __vectorcall rotated(double pitch, double yaw, double roll) const
		{
			return *this * DirectX::XMMatrixRotationRollPitchYaw(static_cast<float>(pitch), static_cast<float>(yaw), static_cast<float>(roll));
		}

		Mat4x4 __vectorcall rotated(const Vec3& pitchYawRoll) const
		{
			return *this * DirectX::XMMatrixRotationRollPitchYaw(static_cast<float>(pitchYawRoll.x), static_cast<float>(pitchYawRoll.y), static_cast<float>(pitchYawRoll.z));
		}

		Mat4x4 __vectorcall rotated(const Quaternion& quaternion) const
		{
			return *this * quaternion.toMatrix();
		}


		bool decompose(Vec3& scale, Quaternion& rotation, Vec3& transform) const;

		double determinant() const;

		Mat4x4 inverse() const;

		Mat4x4 inverse(double& determinant) const;

		Mat4x4 transposed() const;
	};

	template<typename CharType>
	inline
		std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Mat4x4& mat
		)
	{
		return os << CharType('(')
			<< mat.r[0] << CharType(',')
			<< mat.r[1] << CharType(',')
			<< mat.r[2] << CharType(',')
			<< mat.r[3] << CharType(')');
	}

	template<typename CharType>
	inline
		std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Mat4x4& mat
		)
	{
		CharType unused;

		return	is >> unused
			>> mat.r[0] >> unused
			>> mat.r[1] >> unused
			>> mat.r[2] >> unused
			>> mat.r[3] >> unused;
	}
}
