﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"

namespace s3d
{
	/// <summary>
	/// マウス
	/// </summary>
	/// <remarks>
	/// マウス入力の機能を提供します。
	/// </remarks>
	namespace Mouse
	{
		/// <summary>
		/// マウスカーソルがクライアント領域上にあるかを返します。
		/// </summary>
		/// <returns>
		/// マウスカーソルがクライアント領域上にある場合 true, それ以外の場合は false
		/// </returns>
		bool OnClient();

		/// <summary>
		/// クライアント領域が左クリックされたかを返します。
		/// </summary>
		/// <returns>
		/// クライアント領域が左クリックされた場合 true, それ以外の場合は false
		/// </returns>
		bool LeftClicked();

		/// <summary>
		/// クライアント領域が右クリックされたかを返します。
		/// </summary>
		/// <returns>
		/// クライアント領域が右クリックされた場合 true, それ以外の場合は false
		/// </returns>
		bool RightClicked();

		/// <summary>
		/// クライアント座標系でのマウスカーソルの位置を返します。
		/// </summary>
		/// <returns>
		/// クライアント座標系でのマウスカーソルの位置
		/// </returns>
		Point Pos();

		/// <summary>
		/// クライアント座標系での 1 フレーム前のマウスカーソルの位置を返します。
		/// </summary>
		/// <returns>
		/// クライアント座標系での 1 フレーム前のマウスカーソルの位置
		/// </returns>
		Point PreviousPos();

		/// <summary>
		/// クライアント座標系でのマウスカーソルの移動量を返します。
		/// </summary>
		/// <returns>
		/// クライアント座標系でのマウスカーソルの移動量
		/// </returns>
		Point Delta();

		/// <summary>
		/// クライアント座標系での 1 フレーム前と現在のフレームの中間のマウスカーソルの位置を返します。
		/// </summary>
		/// <returns>
		/// クライアント座標系での 1 フレーム前と現在のフレームの中間のマウスカーソルの位置
		/// </returns>
		Point HalfPreviousPos();

		/// <summary>
		/// スクリーン座標系でのマウスカーソルの位置を返します。
		/// </summary>
		/// <returns>
		/// スクリーン座標系でのマウスカーソルの位置
		/// </returns>
		Point ScreenPos();

		/// <summary>
		/// スクリーン座標系での 1 フレーム前のマウスカーソルの位置を返します。
		/// </summary>
		/// <returns>
		/// スクリーン座標系での 1 フレーム前のマウスカーソルの位置
		/// </returns>
		Point PreviousScreenPos();

		/// <summary>
		/// スクリーン座標系でのマウスカーソルの移動量を返します。
		/// </summary>
		/// <returns>
		/// スクリーン座標系でのマウスカーソルの移動量
		/// </returns>
		Point ScreenDelta();

		/// <summary>
		/// クライアント座標系のマウスカーソル位置に適用する座標変換を設定します。
		/// </summary>
		/// <param name="scale">
		/// 適用するスケーリング
		/// </param>
		/// <param name="offset">
		/// 適用するオフセット（ピクセル）
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetTransform(const Mat3x2& matrix);

		/// <summary>
		/// 現在の、クライアント座標系のマウスカーソル位置に適用されている座標変換を返します。
		/// </summary>
		/// <returns>
		/// 適用されているスケーリングとオフセット（ピクセル）のペア
		/// </returns>
		Mat3x2 GetTransform();

		/// <summary>
		/// 垂直方向のマウスホイールの回転量を返します。
		/// </summary>
		/// <returns>
		/// 垂直方向のマウスホイールの回転量
		/// </returns>
		int Wheel();

		/// <summary>
		/// 水平方向のマウスホイールの回転量を返します。
		/// </summary>
		/// <returns>
		/// 水平方向のマウスホイールの回転量
		/// </returns>
		int WheelH();

		/// <summary>
		/// 3D シーンに対するマウスカーソルのレイを取得します。
		/// </summary>
		/// <returns>
		/// 3D シーンに対するマウスカーソルのレイ
		/// </returns>
		const s3d::Ray& Ray();
	}
}
