﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <utility>
# include "Siv3DFwd.hpp"

namespace s3d
{
	enum class FillMode : uint8
	{
		Wireframe	= 2,
		Solid		= 3,
	};

	enum class CullMode : uint8
	{
		None		= 1,
		Front		= 2,
		Back		= 3,
	};

	/// <summary>
	/// ラスタライザーステート
	/// </summary>
	struct RasterizerState
	{
		RasterizerState(
			FillMode _fillMode = FillMode::Solid,
			CullMode _cullMode = CullMode::Back,
			bool _frontCounterClockwise = false,
			int32 _depthBias = 0
			)
			: fillMode(_fillMode)
			, cullMode(_cullMode)
			, frontCounterClockwise(_frontCounterClockwise)
			, depthBias(_depthBias) {}

		using DataType = uint64;

#pragma warning(disable:4201)
		union
		{
			struct
			{
				/// <summary>
				/// 面の塗りつぶし方法
				/// </summary>
				FillMode fillMode : 2;

				/// <summary>
				/// カリングの基準
				/// </summary>
				CullMode cullMode : 2;

				/// <summary>
				/// 頂点が反時計回りの面を表面とするかどうかを示します。 
				/// </summary>
				bool frontCounterClockwise : 4;

				/// <summary>
				/// デプスのバイアス
				/// </summary>
				int32 depthBias;
			};

			DataType _data;
		};
#pragma warning(default:4201)

		bool operator == (const RasterizerState& r) const
		{
			return _data == r._data;
		}

		bool operator != (const RasterizerState& r) const
		{
			return _data != r._data;
		}

		enum class Predefined
		{
			SolidBack,
			SolidFront,
			SolidNone,
			WireframeBack,
			WireframeFront,
			WireframeNone,
			Default2D = SolidNone,
			Default3D = SolidBack,
		};

		RasterizerState(Predefined predefined);

		/// <summary>
		/// ソリッド表示、裏向きの面をカリング
		/// RasterizerState{ FillMode::Solid, CullMode::Back }
		/// </summary>
		const static Predefined SolidBack = Predefined::SolidBack;

		/// <summary>
		/// ソリッド表示、表向きの面をカリング
		/// RasterizerState{ FillMode::Solid, CullMode::Front }
		/// </summary>
		const static Predefined SolidFront = Predefined::SolidFront;

		/// <summary>
		/// ソリッド表示、カリングなし
		/// RasterizerState{ FillMode::Solid, CullMode::None }
		/// </summary>
		const static Predefined SolidNone = Predefined::SolidNone;

		/// <summary>
		/// ワイヤフレーム表示、裏向きの面をカリング
		/// RasterizerState{ FillMode::Wireframe, CullMode::Back }
		/// </summary>
		const static Predefined WireframeBack = Predefined::WireframeBack;

		/// <summary>
		/// ワイヤフレーム表示、表向きの面をカリング
		/// RasterizerState{ FillMode::Wireframe, CullMode::Front }
		/// </summary>
		const static Predefined WireframeFront = Predefined::WireframeFront;

		/// <summary>
		/// ワイヤフレーム表示、カリングなし
		/// RasterizerState{ FillMode::Wireframe, CullMode::None }
		/// </summary>
		const static Predefined WireframeNone = Predefined::WireframeNone;

		/// <summary>
		/// 2D 描画時のデフォルト
		/// RasterizerState::SolidNone
		/// </summary>
		const static Predefined Default2D = Predefined::Default2D;

		/// <summary>
		/// 3D 描画時のデフォルト
		/// RasterizerState::SolidBack
		/// </summary>
		const static Predefined Default3D = Predefined::Default3D;
	};

	static_assert(sizeof(RasterizerState) == sizeof(RasterizerState::DataType), "");
}

template<>
struct std::hash<s3d::RasterizerState>
{
	size_t operator()(const s3d::RasterizerState& keyVal) const
	{
		return std::hash<s3d::RasterizerState::DataType>()(keyVal._data);
	}
};
