﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivString.hpp"
# include "SivArray.hpp"

namespace s3d
{
	/// <summary>
	/// 録音のサンプリングレート
	/// </summary>
	namespace RecorderFormat
	{
		enum Type
		{
			/// <summary>
			/// ステレオ 11,025 Hz
			/// </summary>
			S11025 = 0x00000008,

			/// <summary>
			/// ステレオ 22,050 Hz
			/// </summary>
			S22050 = 0x00000080,

			/// <summary>
			/// ステレオ 44,100 Hz
			/// </summary>
			S44100 = 0x00000800,

			/// <summary>
			/// ステレオ 48,000 Hz
			/// </summary>
			S48000 = 0x00008000,

			/// <summary>
			/// ステレオ 96,000 Hz
			/// </summary>
			S96000 = 0x00080000,

			/// <summary>
			/// デフォルト [ステレオ 48,000 Hz]
			/// </summary>
			Default = S44100,
		};
	}

	/// <summary>
	/// 録音デバイスの情報
	/// </summary>
	struct RecorderInfo
	{
		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		RecorderInfo() = default;

		/// <summary>
		/// エンジン内部で使用するコンストラクタ
		/// </summary>
		RecorderInfo(const String& _name, int32 _format, int32 _channels, uint32 _deviceID)
			: name(_name)
			, format(_format)
			, channels(_channels)
			, deviceID(_deviceID) {}

		/// <summary>
		/// デバイス名
		/// </summary>
		/// <remarks>
		/// 文字数によっては名前が途中で切られている場合があります。
		/// </remarks>
		String name;

		/// <summary>
		/// 対応した録音フォーマット（RecorderFormat::Type の組み合わせ）
		/// </summary>
		int32 format = 0;

		/// <summary>
		/// 録音できるチャンネル数
		/// </summary>
		int32 channels = 0;

		/// <summary>
		/// 
		/// </summary>
		uint32 deviceID = 0;
	};

	/// <summary>
	/// 録音
	/// </summary>
	/// <remarks>
	/// 録音に関する機能を提供します。
	/// </remarks>
	namespace Recorder
	{
		/// <summary>
		/// PC に接続されている録音デバイスの一覧を取得します。
		/// </summary>
		/// <returns>
		/// PC に接続されている録音デバイスの一覧
		/// </returns>
		Array<RecorderInfo> Enumerate();

		/// <summary>
		/// 使用する録音デバイスを変更します。
		/// </summary>
		/// <param name="id">
		/// デバイス ID
		/// </param>
		/// <remarks>
		/// この関数が呼ばれると、現在の録音は終了します。
		/// </remarks>
		/// <returns>
		/// 録音デバイスの変更に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool SetDevice(uint32 id);

		/// <summary>
		/// 録音用のバッファのフォーマットとサイズをサンプル数で設定します。
		/// </summary>
		/// <param name="format">
		/// 録音のサンプリングレート
		/// </param>
		/// <param name="sample">
		/// 録音を保存するバッファのサンプル数
		/// </param>
		/// <returns>
		/// バッファの設定に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool SetBufferLengthBySample(RecorderFormat::Type format, uint32 sample);

		/// <summary>
		/// 録音用のバッファのフォーマットとサイズを秒で設定します。
		/// </summary>
		/// <param name="format">
		/// 録音のサンプリングレート
		/// </param>
		/// <param name="timeSec">
		/// 録音を保存するバッファの長さ（秒）
		/// </param>
		/// <returns>
		/// バッファの設定に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool SetBufferLengthBySec(RecorderFormat::Type format, double timeSec);

		/// <summary>
		/// 録音を開始します。
		/// </summary>
		/// <returns>
		/// 録音の開始に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Start();

		/// <summary>
		/// 録音を終了します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void Stop();

		/// <summary>
		/// 録音しているバッファの現在の位置を返します。
		/// </summary>
		/// <returns>
		/// 録音しているバッファの現在の位置
		/// </returns>
		uint32 GetPosSample();

		__declspec(deprecated("** Recorder::GetPos() is a deprecated function. Use Recorder::GetPosSample() instead. **"))
			bool GetPos();

		/// <summary>
		/// 録音を保存しているバッファにアクセスします。
		/// </summary>
		/// <remarks>
		/// 録音中は、GetPos() が返す位置以降のサンプルにアクセスしてはいけません。
		/// </remarks>
		/// <returns>
		/// 録音を保存しているバッファ
		/// </returns>
		const Wave& GetWave();

		/// <summary>
		/// 録音がバッファの終端に達して終了しているかを返します。
		/// </summary>
		/// <returns>
		/// 録音がバッファの終端に達して終了している場合 true, それ以外の場合は false
		/// </returns>
		bool IsEnded();

		__declspec(deprecated("** Recorder::IsEnd() is a deprecated function. Use Recorder::IsEnded() instead. **"))
		bool IsEnd();

		/// <summary>
		/// 録音をバッファの最初から再開します。
		/// </summary>
		/// <returns>
		/// 録音の再開に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Restart();

		/// <summary>
		/// 録音している音声の直近の最大振幅を返します。
		/// </summary>
		/// <param name="sec">
		/// 計測する直近の波形の範囲（秒）
		/// </param>
		/// <returns>
		/// 音声の直近の最大振幅
		/// </returns>
		double GetMaxAmplitude(double sec = 0.02);

		/// <summary>
		/// 録音している音声の直近の平均振幅を返します。
		/// </summary>
		/// <param name="sec">
		/// 計測する直近の波形の範囲（秒）
		/// </param>
		/// <returns>
		/// 音声の直近の平均振幅
		/// </returns>
		double GetAverageAmplitude(double sec = 0.02);
	}
}
