﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "SivArray.hpp"
# include "SivString.hpp"
# include "SivOptional.hpp"

namespace s3d
{
	/// <summary>
	/// シリアル通信
	/// </summary>
	class Serial
	{
	private:

		class CSerial;

		std::shared_ptr<CSerial> pSerial;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Serial();

		/// <summary>
		/// シリアルポートを開きます。
		/// </summary>
		/// <param name="port">
		/// COM ポート番号
		/// </param>
		/// <param name="baudRate">
		/// 転送レート
		/// </param>
		explicit Serial(int32 port, int32 baudRate = 9600);

		/// <summary>
		/// デストラクタ
		/// </summary>
		~Serial();

		/// <summary>
		/// シリアルポートを開きます。
		/// </summary>
		/// <param name="port">
		/// COM ポート番号
		/// </param>
		/// <param name="baudRate">
		/// 転送レート
		/// </param>
		/// <returns>
		/// ポートのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool setup(int32 port, int32 baudRate = 9600);

		/// <summary>
		/// シリアルデバイスに接続されているかを返します。
		/// </summary>
		/// <returns>
		/// 接続されている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// シリアルデバイスに接続されているかを返します。
		/// </summary>
		/// <returns>
		/// 接続されている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const;

		/// <summary>
		/// シリアルポートから読み込めるデータのサイズを取得します。
		/// </summary>
		/// <returns>
		/// シリアルポートから読み込めるデータのサイズ（バイト）
		/// </returns>
		int available();

		/// <summary>
		/// シリアルの読み込みバッファの内容を消去します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void flushInput();

		/// <summary>
		/// シリアルの書き込みバッファの内容を消去します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void flushOutput();

		/// <summary>
		/// シリアルのバッファの内容を消去します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void flush();

		/// <summary>
		/// シリアルデバイスから 1 バイトのデータを読み込みます。
		/// </summary>
		/// <returns>
		/// 読み込みに成功した場合はそのデータの Optional, 失敗した場合は none
		/// </returns>
		Optional<uint8> readByte();

		/// <summary>
		/// シリアルデバイスからデータを読み込みます。
		/// </summary>
		/// <returns>
		/// 読み込みに成功した場合はそのデータの Optional, 失敗した場合は none
		/// </returns>
		Optional<Array<uint8>> readBytes();

		/// <summary>
		/// シリアルデバイスから 1 バイトのデータを読み込みます。
		/// </summary>
		/// <param name="byte">
		/// 読み込み先
		/// </param>
		/// <returns>
		/// 読み込みに成功した場合は true, 失敗した場合は false
		/// </returns>
		bool readByte(uint8& byte);

		/// <summary>
		/// シリアルデバイスからデータを読み込みます。
		/// </summary>
		/// <param name="bytes">
		/// 読み込み先
		/// </param>
		/// <returns>
		/// 読み込みに成功した場合は true, 失敗した場合は false
		/// </returns>
		bool readBytes(Array<uint8>& bytes);

		/// <summary>
		/// シリアルデバイスに 1 バイトのデータを書き込みます。
		/// </summary>
		/// <param name="byte">
		/// 書き込むデータ
		/// </param>
		/// <returns>
		/// 実際に書き込んだサイズ
		/// </returns>
		int writeByte(uint8 byte);

		/// <summary>
		/// シリアルデバイスにデータを書き込みます。
		/// </summary>
		/// <param name="byte">
		/// 書き込むデータの先頭ポインタ
		/// </param>
		/// <param name="size">
		/// 書き込むデータのサイズ
		/// </param>
		/// <returns>
		/// 実際に書き込んだサイズ
		/// </returns>
		int writeBytes(const uint8* bytes, size_t size);

		/// <summary>
		/// シリアルデバイスに文字列を書き込みます。
		/// </summary>
		/// <param name="str">
		/// 書き込む文字列
		/// </param>
		/// <remarks>
		/// 文字列は UTF-8 エンコードされます。
		/// </remarks>
		/// <returns>
		/// 実際に書き込んだサイズ
		/// </returns>
		int write(const String& str);

		/// <summary>
		/// シリアルデバイスに文字列と改行を書き込みます。
		/// </summary>
		/// <param name="str">
		/// 書き込む文字列
		/// </param>
		/// <remarks>
		/// 文字列は UTF-8 エンコードされます。
		/// </remarks>
		/// <returns>
		/// 実際に書き込んだサイズ
		/// </returns>
		int writeln(const String& str);

		/// <summary>
		/// シリアルデバイスとの接続を終了します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void close();
	};

	/// <summary>
	/// シリアルデバイスの情報
	/// </summary>
	struct SerialDeveiceInfo
	{
		/// <summary>
		/// シリアルデバイスの名前
		/// </summary>
		String name;

		/// <summary>
		/// COM ポート番号
		/// </summary>
		int32 port = 0;
	};

	/// <summary>
	/// シリアル通信
	/// </summary>
	/// <remarks>
	/// シリアル通信の機能を提供します。
	/// </remarks>
	namespace SerialManager
	{
		/// <summary>
		/// PC に接続されているシリアルデバイスの一覧を取得します。
		/// </summary>
		/// <returns>
		/// PC に接続されているシリアルデバイスの一覧
		/// </returns>
		Array<SerialDeveiceInfo> Enumerate();
	}
}
