﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <iostream>
# include "SivPropertyMacro.hpp"
# include "SivPoint.hpp"

namespace s3d
{
	/// <summary>
	/// 2 次元ベクトル
	/// </summary>
	template<typename Type>
	struct Vector2D
	{
		using value_type = Type;

		Type x, y;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Vector2D() = default;
		
		template<typename U, typename V>
		Vector2D(U _x, V _y) : x(static_cast<Type>(_x)), y(static_cast<Type>(_y)) {}

		Vector2D(Type _x, Type _y) : x(_x), y(_y) {}

		Vector2D(const Point& v) : x(static_cast<Type>(v.x)), y(static_cast<Type>(v.y)) {}

		template<typename U>
		Vector2D(const Vector2D<U>& v) : x(static_cast<Type>(v.x)), y(static_cast<Type>(v.y)) {}

		Vector2D operator + () const;

		Vector2D operator - () const;

		Vector2D operator + (const Vector2D& v) const;

		Vector2D operator - (const Vector2D& v) const;

		Vector2D operator * (Type s) const;

		Vector2D operator * (const Vector2D& v) const;

		Vector2D operator / (Type s) const;

		Vector2D& operator += (const Vector2D& v);

		Vector2D& operator -= (const Vector2D& v);

		Vector2D& operator *= (Type s);

		Vector2D& operator /= (Type s);

		bool operator == (const Vector2D& v) const;

		bool operator != (const Vector2D& v) const;

		/// <summary>
		/// Vector2D{ x, x }
		/// </summary>
		Property_Get(Vector2D, xx) const;

		/// <summary>
		/// Vector2D{ x, y }
		/// </summary>
		Property_Get(Vector2D, xy) const;

		/// <summary>
		/// Vector2D{ y, x }
		/// </summary>
		Property_Get(Vector2D, yx) const;

		/// <summary>
		/// Vector2D{ y, y }
		/// </summary>
		Property_Get(Vector2D, yy) const;

		Point asPoint() const
		{
			return{ static_cast<int>(x), static_cast<int>(y) };
		}

		Vector2D& set(Type _x, Type _y) { x = _x; y = _y; return *this; }

		Vector2D& set(const Vector2D& v) { return *this = v; }

		Vector2D movedBy(Type _x, Type _y) const { return{ x + _x, y + _y }; }

		Vector2D movedBy(const Vector2D& v) const { return{ x + v.x, y + v.y }; }

		Vector2D& moveBy(Type _x, Type _y) { x += _x; y += _y; return *this; }

		Vector2D& moveBy(const Vector2D& v) { return *this += v; }

		template<typename Shape>
		bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		Property_Get(bool, isZero) const;

		Type length() const;

		Type lengthSq() const;

		Type lengthInv() const;

		Type distanceFrom(const Vector2D& v) const;

		Type distanceFromSq(const Vector2D& v) const;

		Type dot(const Vector2D& v) const;

		Type cross(const Vector2D& v) const;

		Vector2D normalized() const;

		Vector2D& normalize();

		Vector2D rotated(double angle) const;

		Vector2D& rotate(double angle);

		Vector2D projection(const Vector2D& onto) const;

		const Vector2D& draw(const Color& color = Palette::White) const;

		/// <summary>
		/// Vector2D{ 0, 0 }
		/// </summary>
		const static Vector2D Zero;

		/// <summary>
		/// Vector2D{ 1, 1 }
		/// </summary>
		const static Vector2D One;

		/// <summary>
		/// Vector2D{ 1, 0 }
		/// </summary>
		const static Vector2D UnitX;

		/// <summary>
		/// Vector2D{ 0, 1 }
		/// </summary>
		const static Vector2D UnitY;

		/// <summary>
		/// Vector2D{ -1, 0 }
		/// </summary>
		const static Vector2D Left;

		/// <summary>
		/// Vector2D{ 1, 0 }
		/// </summary>
		const static Vector2D Right;

		/// <summary>
		/// Vector2D{ 0, -1 }
		/// </summary>
		const static Vector2D Up;

		/// <summary>
		/// Vector2D{ 0, 1 }
		/// </summary>
		const static Vector2D Down;
	};

	/// <summary>
	/// 出力ストリームに 2 次元ベクトルを渡します。
	/// </summary>
	/// <param name="os">
	/// 出力ストリーム
	/// </param>
	/// <param name="v">
	/// 2 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の出力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Vector2D<Type>& v
		)
	{
		return os << CharType('(') << v.x << CharType(',') << v.y << CharType(')');
	}

	/// <summary>
	/// 入力ストリームに 2 次元ベクトルを渡します。
	/// </summary>
	/// <param name="is">
	/// 入力ストリーム
	/// </param>
	/// <param name="v">
	/// 2 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の入力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Vector2D<Type>& v
		)
	{
		CharType unused;
		return is >> unused >> v.x >> unused >> v.y >> unused;
	}

	using Float2	= Vector2D<float>;
	using Vec2		= Vector2D<double>;

	template<typename Type, typename U>
	inline Vector2D<Type> operator * (U s, const Vector2D<Type>& v)
	{
		return v*s;
	}

	template<typename Type>
	inline Vector2D<Type> operator + (const Point& p, const Vector2D<Type>& v)
	{
		return v + p;
	}

	template<typename Type>
	inline Vector2D<Type> operator - (const Point& p, const Vector2D<Type>& v)
	{
		return Vector2D<Type>(p)-v;
	}
}
