﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <iostream>
# include "SivPropertyMacro.hpp"
# include "SivPointVector.hpp"

namespace s3d
{
	/// <summary>
	/// 3 次元ベクトル
	/// </summary>
	template<typename Type>
	struct Vector3D
	{
		using value_type = Type;

		Type x, y, z;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Vector3D() = default;

		template<typename U, typename V, typename W>
		Vector3D(U _x, V _y, W _z) : x(static_cast<Type>(_x)), y(static_cast<Type>(_y)), z(static_cast<Type>(_z)) {}

		Vector3D(Type _x, Type _y, Type _z) : x(_x), y(_y), z(_z) {}

		template<typename U>
		Vector3D(const Vector2D<U>& xy, Type _z) : x(static_cast<Type>(xy.x)), y(static_cast<Type>(xy.y)), z(_z) {}

		template<typename U>
		Vector3D(Type _x, const Vector2D<U>& yz) : x(_x), y(static_cast<Type>(yz.x)), z(static_cast<Type>(yz.y)) {}

		template<typename U>
		Vector3D(const Vector3D<U>& v) : x(static_cast<Type>(v.x)), y(static_cast<Type>(v.y)), z(static_cast<Type>(v.z)) {}

		Vector3D operator + () const;

		Vector3D operator - () const;

		Vector3D operator + (const Vector3D& v) const;

		Vector3D operator - (const Vector3D& v) const;

		Vector3D operator * (Type s) const;

		Vector3D operator * (const Vector3D& v) const;

		Vector3D operator / (Type s) const;

		Vector3D& operator += (const Vector3D& v);

		Vector3D& operator -= (const Vector3D& v);

		Vector3D& operator *= (Type s);

		Vector3D& operator /= (Type s);

		/// <summary>
		/// Vector2D{ x, y }
		/// </summary>
		Property_Get(Vector2D<Type>, xy) const;

		/// <summary>
		/// Vector2D{ x, z }
		/// </summary>
		Property_Get(Vector2D<Type>, xz) const;

		/// <summary>
		/// Vector2D{ y, z }
		/// </summary>
		Property_Get(Vector2D<Type>, yz) const;

		Vector3D& set(Type _x, Type _y, Type _z) { x = _x; y = _y; z = _z; return *this; }

		Vector3D& set(const Vector3D& v) { return *this = v; }

		Vector3D movedBy(Type _x, Type _y, Type _z) const { return{ x + _x, y + _y, z + _z }; }

		Vector3D movedBy(const Vector3D& v) const { return{ x + v.x, y + v.y, z + v.z }; }

		Vector3D& moveBy(Type _x, Type _y, Type _z) { x += _x; y += _y; z += _z; return *this; }

		Vector3D& moveBy(const Vector3D& v) { return *this += v; }

		Property_Get(bool, isZero) const;

		Type length() const;

		Type lengthSq() const;

		Type lengthInv() const;

		Type distanceFrom(const Vector3D& v) const;

		Type distanceFromSq(const Vector3D& v) const;

		Type dot(const Vector3D& v) const;

		Vector3D cross(const Vector3D& v) const;

		Vector3D normalized() const;

		Vector3D& normalize();

		template<typename Primitive> bool intersects(const Primitive& primitive) const
		{
			return Geometry3D::Intersect(*this, primitive);
		}

		/// <summary>
		/// Vector3D{ 0, 0, 0 }
		/// </summary>
		const static Vector3D Zero;

		/// <summary>
		/// Vector3D{ 1, 1, 1 }
		/// </summary>
		const static Vector3D One;

		/// <summary>
		/// Vector3D{ 1, 0, 0 }
		/// </summary>
		const static Vector3D UnitX;

		/// <summary>
		/// Vector3D{ 0, 1, 0 }
		/// </summary>
		const static Vector3D UnitY;

		/// <summary>
		/// Vector3D{ 0, 0, 1 }
		/// </summary>
		const static Vector3D UnitZ;

		/// <summary>
		/// Vector3D{ -1, 0, 0 }
		/// </summary>
		const static Vector3D Left;

		/// <summary>
		/// Vector3D{ 1, 0, 0 }
		/// </summary>
		const static Vector3D Right;

		/// <summary>
		/// Vector3D{ 0, 1, 0 }
		/// </summary>
		const static Vector3D Up;

		/// <summary>
		/// Vector3D{ 0, -1, 0 }
		/// </summary>
		const static Vector3D Down;

		/// <summary>
		/// Vector3D{ 0, 0, 1 }
		/// </summary>
		const static Vector3D Forward;

		/// <summary>
		/// Vector3D{ 0, 0, -1 }
		/// </summary>
		const static Vector3D Backward;
	};

	/// <summary>
	/// 出力ストリームに 3 次元ベクトルを渡します。
	/// </summary>
	/// <param name="os">
	/// 出力ストリーム
	/// </param>
	/// <param name="v">
	/// 3 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の出力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
		std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Vector3D<Type>& v
		)
	{
		return os << CharType('(') << v.x << CharType(',') << v.y << CharType(',') << v.z << CharType(')');
	}

	/// <summary>
	/// 入力ストリームに 3 次元ベクトルを渡します。
	/// </summary>
	/// <param name="is">
	/// 入力ストリーム
	/// </param>
	/// <param name="v">
	/// 3 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の入力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
		std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Vector3D<Type>& v
		)
	{
		CharType unused;
		return is >> unused >> v.x >> unused >> v.y >> unused >> v.z >> unused;
	}

	using Float3	= Vector3D<float>;
	using Vec3		= Vector3D<double>;

	template<typename Type, typename U>
	inline Vector3D<Type> operator * (U s, const Vector3D<Type>& v)
	{
		return v*s;
	}
}
