﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <iostream>
# include "SivPropertyMacro.hpp"

namespace s3d
{
	/// <summary>
	/// 4 次元ベクトル
	/// </summary>
	template<typename Type>
	struct Vector4D
	{
		using value_type = Type;

		Type x, y, z, w;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Vector4D() = default;

		template<typename U, typename V, typename W, typename X>
		Vector4D(U _x, V _y, W _z, X _w) : x(static_cast<Type>(_x)), y(static_cast<Type>(_y)), z(static_cast<Type>(_z)), w(static_cast<Type>(_w)){}

		Vector4D(Type _x, Type _y, Type _z, Type _w) : x(_x), y(_y), z(_z), w(_w){}

		template<typename U>
		Vector4D(const Vector2D<U>& xy, Type _z, Type _w) : x(static_cast<Type>(xy.x)), y(static_cast<Type>(xy.y)), z(_z), w(_w){}

		template<typename U>
		Vector4D(Type _x, Type _y, const Vector2D<U>& zw) : x(_x), y(_y), z(static_cast<Type>(zw.x)), w(static_cast<Type>(zw.y)){}

		template<typename U, typename V>
		Vector4D(const Vector2D<U>& xy, const Vector2D<V>& zw) : x(static_cast<Type>(xy.x)), y(static_cast<Type>(xy.y)), z(static_cast<Type>(zw.x)), w(static_cast<Type>(zw.y)){}

		template<typename U>
		Vector4D(const Vector3D<U>& xyz, Type _w) : x(static_cast<Type>(xyz.x)), y(static_cast<Type>(xyz.y)), z(static_cast<Type>(xyz.z)), w(_w){}

		template<typename U>
		Vector4D(Type _x, const Vector3D<U>& yzw) : x(_x), y(static_cast<Type>(yzw.y)), z(static_cast<Type>(yzw.y)), w(static_cast<Type>(yzw.z)){}

		template<typename U>
		Vector4D(const Vector4D<U>& v) : x(static_cast<Type>(v.x)), y(static_cast<Type>(v.y)), z(static_cast<Type>(v.z)), w(static_cast<Type>(v.w)){}

		Vector4D operator + () const;

		Vector4D operator - () const;

		Vector4D operator + (const Vector4D& v) const;

		Vector4D operator - (const Vector4D& v) const;

		Vector4D operator * (Type s) const;

		Vector4D operator * (const Vector4D& v) const;

		Vector4D operator / (Type s) const;

		Vector4D& operator += (const Vector4D& v);

		Vector4D& operator -= (const Vector4D& v);

		Vector4D& operator *= (Type s);

		Vector4D& operator /= (Type s);

		/// <summary>
		/// Vector2D{ x, y }
		/// </summary>
		Property_Get(Vector2D<Type>, xy) const;

		/// <summary>
		/// Vector2D{ y, z }
		/// </summary>
		Property_Get(Vector2D<Type>, yz) const;

		/// <summary>
		/// Vector2D{ z, w }
		/// </summary>
		Property_Get(Vector2D<Type>, zw) const;

		/// <summary>
		/// Vector3D{ x, y, z }
		/// </summary>
		Property_Get(Vector3D<Type>, xyz) const;

		/// <summary>
		/// Vector3D{ y, z, w }
		/// </summary>
		Property_Get(Vector3D<Type>, yzw) const;

		Vector4D& set(Type _x, Type _y, Type _z, Type _w) { x = _x; y = _y; z = _z; w = _w; return *this; }

		Vector4D& set(const Vector4D& v) { return *this = v; }

		Vector4D movedBy(Type _x, Type _y, Type _z, Type _w) const { return{ x + _x, y + _y, z + _z, w + _w }; }

		Vector4D movedBy(const Vector4D& v) const { return{ x + v.x, y + v.y, z + v.z, w + v.w }; }

		Vector4D& moveBy(Type _x, Type _y, Type _z, Type _w) { x += _x; y += _y; z += _z; w += _w; return *this; }

		Vector4D& moveBy(const Vector4D& v) { return *this += v; }

		Property_Get(bool, isZero) const;

		Type length() const;

		Type lengthSq() const;

		Type lengthInv() const;

		Type distanceFrom(const Vector4D& v) const;

		Type distanceFromSq(const Vector4D& v) const;

		Type dot(const Vector4D& v) const;

		Vector4D normalized() const;

		Vector4D& normalize();

		/// <summary>
		/// Vector4D{ 0, 0, 0, 0 }
		/// </summary>
		const static Vector4D Zero;

		/// <summary>
		/// Vector4D{ 1, 1, 1, 1 }
		/// </summary>
		const static Vector4D One;

		/// <summary>
		/// Vector4D{ 1, 0, 0, 0 }
		/// </summary>
		const static Vector4D UnitX;

		/// <summary>
		/// Vector4D{ 0, 1, 0, 0 }
		/// </summary>
		const static Vector4D UnitY;

		/// <summary>
		/// Vector4D{ 0, 0, 1, 0 }
		/// </summary>
		const static Vector4D UnitZ;

		/// <summary>
		/// Vector4D{ 0, 0, 0, 1 }
		/// </summary>
		const static Vector4D UnitW;
	};

	/// <summary>
	/// 出力ストリームに 4 次元ベクトルを渡します。
	/// </summary>
	/// <param name="os">
	/// 出力ストリーム
	/// </param>
	/// <param name="v">
	/// 4 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の出力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
		std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Vector4D<Type>& v
		)
	{
		return os << CharType('(') << v.x << CharType(',') << v.y << CharType(',') << v.z << CharType(',') << v.w << CharType(')');
	}

	/// <summary>
	/// 入力ストリームに 4 次元ベクトルを渡します。
	/// </summary>
	/// <param name="is">
	/// 入力ストリーム
	/// </param>
	/// <param name="v">
	/// 4 次元ベクトル
	/// </param>
	/// <returns>
	/// 渡した後の入力ストリーム
	/// </returns>
	template<typename CharType, typename Type>
	inline
		std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Vector4D<Type>& v
		)
	{
		CharType unused;
		return is >> unused >> v.x >> unused >> v.y >> unused >> v.z >> unused >> v.w >> unused;
	}

	using Float4	= Vector4D<float>;
	using Vec4		= Vector4D<double>;

	template<typename Type, typename U>
	inline Vector4D<Type> operator * (U s, const Vector4D<Type>& v)
	{
		return v*s;
	}
}
