﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2015 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivOptional.hpp"
# include "SivString.hpp"
# include "SivWaveSample.hpp"
# include "SivMath.hpp"
# include "SivMathConstants.hpp"

namespace s3d
{
	/// <summary>
	/// オーディオファイルのフォーマット
	/// </summary>
	enum class AudioFormat
	{
		/// <summary>
		/// 
		/// </summary>
		Unknown,

		/// <summary>
		/// WAVE
		/// </summary>
		WAV,

		/// <summary>
		/// OggVorbis
		/// </summary>
		OggVorbis,

		/// <summary>
		/// MP3
		/// </summary>
		MP3,

		/// <summary>
		/// AAC(M4A)
		/// </summary>
		AAC,

		/// <summary>
		/// WMA
		/// </summary>
		WMA,

		//Aiff,

		//Flac,

		/// <summary>
		/// 
		/// </summary>
		Unspecified = Unknown,
	};

	/// <summary>
	/// オーディオの情報
	/// </summary>
	struct AudioProperty
	{
		/// <summary>
		/// 
		/// </summary>
		AudioFormat format;

		/// <summary>
		/// 再生時間（ミリ秒）
		/// </summary>
		uint32 durationMillisec;

		/// <summary>
		/// ビットレート
		/// </summary>
		uint32 encodingBitrate;

		/// <summary>
		/// タイトル
		/// </summary>
		String title;

		/// <summary>
		/// アーティスト
		/// </summary>
		String artist;

		/// <summary>
		/// アルバムタイトル
		/// </summary>
		String albumTitle;

		/// <summary>
		/// アルバムアーティスト
		/// </summary>
		String albumArtist;

		/// <summary>
		/// 
		/// </summary>
		Optional<unsigned> year;

		/// <summary>
		/// 
		/// </summary>
		Optional<unsigned> trackNumber;
	};

	/// <summary>
	/// WAVE の保存形式
	/// </summary>
	enum class WAVEFormat
	{
		/// <summary>
		/// 8bit ステレオ
		/// </summary>
		Stereo8U,

		/// <summary>
		/// 16bit ステレオ
		/// </summary>
		Stereo16S,

		/// <summary>
		/// 32bit float ステレオ
		/// </summary>
		Stereo32F,

		/// <summary>
		/// 
		/// </summary>
		Default = Stereo16S,
	};

	/// <summary>
	/// MP3 オーディオの圧縮品質
	/// </summary>
	enum class MP3Quality
	{
		/// <summary>
		/// 32kbps
		/// </summary>
		BR32kbps,

		/// <summary>
		/// 40kbps
		/// </summary>
		BR40kbps,

		/// <summary>
		/// 48kbps
		/// </summary>
		BR48kbps,

		/// <summary>
		/// 56kbps
		/// </summary>
		BR56kbps,

		/// <summary>
		/// 64kbps
		/// </summary>
		BR64kbps,

		/// <summary>
		/// 80kbps
		/// </summary>
		BR80kbps,

		/// <summary>
		/// 96kbps
		/// </summary>
		BR96kbps,

		/// <summary>
		/// 112kbps
		/// </summary>
		BR112kbps,

		/// <summary>
		/// 128kbps
		/// </summary>
		BR128kbps,

		/// <summary>
		/// 160kbps
		/// </summary>
		BR160kbps,

		/// <summary>
		/// 192kbps
		/// </summary>
		BR192kbps,

		/// <summary>
		/// 224kbps
		/// </summary>
		BR224kbps,

		/// <summary>
		/// 256kbps
		/// </summary>
		BR256kbps,

		/// <summary>
		/// 320kbps
		/// </summary>
		BR320kbps,

		/// <summary>
		/// 
		/// </summary>
		Default = BR192kbps,
	};

	/// <summary>
	/// AAC オーディオの圧縮品質
	/// </summary>
	enum class AACQuality
	{
		/// <summary>
		/// 96kbps
		/// </summary>
		BR96kbps,

		/// <summary>
		/// 128kbps
		/// </summary>
		BR128kbps,

		/// <summary>
		/// 160kbps
		/// </summary>
		BR160kbps,

		/// <summary>
		/// 192kbps
		/// </summary>
		BR192kbps,

		/// <summary>
		/// 
		/// </summary>
		Default = BR160kbps,
	};

	/// <summary>
	/// 波形
	/// </summary>
	class Wave
	{
	private:

		Array<WaveSample> m_data;

		uint32 m_samplingRate = DefaultSamplingRate;

	public:

		using iterator			= Array<WaveSample>::iterator;
		using const_iterator	= Array<WaveSample>::const_iterator;
		using reverse_iterator	= Array<WaveSample>::reverse_iterator;	
		using const_reverse_iterator = Array<WaveSample>::const_reverse_iterator;

		enum
		{
			/// <summary>
			/// 
			/// </summary>
			MinSamlingRate = 4000,

			/// <summary>
			/// 
			/// </summary>
			DefaultSamplingRate = 44100,

			/// <summary>
			/// 
			/// </summary>
			MaxSamplingRate = 192000,
		};

		/// <summary>
		/// 
		/// </summary>
		Wave();

		/// <summary>
		/// 
		/// </summary>
		~Wave();

		explicit Wave(size_t sample);

		Wave(size_t sample, uint32 samplingRate);

		Wave(size_t sample, uint32 samplingRate, double hz);

		explicit Wave(const FilePath& path);

		explicit Wave(IReader&& reader, AudioFormat format = AudioFormat::Unspecified);

		Wave(const Wave& wave);

		Wave(Wave && wave);

		Wave& operator =(const Wave& wave);

		Wave& operator =(Wave && wave);

		void clear();

		void release()
		{
			clear();

			shrink_to_fit();
		}

		void shrink_to_fit()
		{
			m_data.shrink_to_fit();
		}

		void fillZero();

		unsigned setSamplingRate(size_t samplingRate);

		explicit operator bool() const
		{
			return !isEmpty;
		}

		void resize(size_t sample);

		WaveSample& at(size_t pos);

		const WaveSample& at(size_t pos) const;

		WaveSample& operator [](size_t pos)
		{
			return m_data[pos];
		}

		const WaveSample& operator [](size_t pos) const
		{
			return m_data[pos];
		}

		void* data()
		{
			return &m_data[0];
		}

		const void* data() const
		{
			return &m_data[0];
		}

		iterator begin() { return m_data.begin(); }

		const_iterator begin() const { return m_data.begin(); }

		const_iterator cbegin() const { return m_data.cbegin(); }

		iterator end() { return m_data.end(); }

		const_iterator end() const { return m_data.end(); }

		const_iterator cend() const { return m_data.cend(); }

		reverse_iterator rbegin() { return m_data.rbegin(); }

		const_reverse_iterator rbegin() const { return m_data.rbegin(); }

		const_reverse_iterator crbegin() const { return m_data.crbegin(); }

		reverse_iterator rend() { return m_data.rend(); }

		const_reverse_iterator rend() const { return m_data.rend(); }

		const_reverse_iterator crend() const { return m_data.crend(); }

		void swap(Wave& wav)
		{
			m_data.swap(wav.m_data);
			
			std::swap(m_samplingRate, wav.m_samplingRate);
		}

		Wave clone() const
		{
			return *this;
		}

		/// <summary>
		/// 
		/// </summary>
		Property_Get(double, lengthSec) const { return static_cast<double>(lengthSample) / m_samplingRate; }

		/// <summary>
		/// 
		/// </summary>
		Property_Get(uint32, lengthSample) const { return m_data.size(); }

		/// <summary>
		/// 
		/// </summary>
		Property_Get(uint32, samplingRate) const { return m_samplingRate; }
		
		/// <summary>
		/// 
		/// </summary>
		Property_Get(uint32, memorySize) const { return lengthSample * sizeof(WaveSample); }

		/// <summary>
		/// 
		/// </summary>
		Property_Get(bool, isEmpty) const { return m_data.empty(); }

		bool save(const FilePath& path, AudioFormat format = AudioFormat::Unspecified) const;

		bool saveWAVE(const FilePath& path, WAVEFormat format = WAVEFormat::Default) const;

		/// <summary>
		/// 
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <param name="quality">
		/// 圧縮品質 [0,100]
		/// </param>
		/// <returns>
		/// 
		/// </returns>
		bool saveOggVorbis(const FilePath& path, int32 quality = 60) const;

		bool saveMP3(const FilePath& path, MP3Quality quality = MP3Quality::Default) const;

		bool saveAAC(const FilePath& path, AACQuality quality = AACQuality::Default) const;

		MemoryReader encode(AudioFormat format) const;

		MemoryReader encodeOggVorbis(int32 quality = 60) const;

		MemoryReader encodeMP3(MP3Quality quality = MP3Quality::Default) const;

		MemoryReader encodeAAC(AACQuality quality = AACQuality::Default) const;

		Wave& swapChannel();

		Wave& reverse();

		Wave& convertTo32FMono();
	};

	/// <summary>
	/// 
	/// </summary>
	namespace Audio
	{
		AudioFormat GetFormat(const FilePath& path);

		AudioFormat GetFormat(const IReader& reader);

		AudioFormat GetFormatFromFileName(const FilePath& path);

		Optional<AudioProperty> GetProperty(const FilePath& path);

		bool MP3EncoderAvailable();

		bool AACEncoderAvailable();

		bool WMAEncoderAvailable();
	}

	class WaveStretcher
	{
	private:

		class CWaveStretcher;

		std::shared_ptr<CWaveStretcher> m_stretcher;

	public:

		WaveStretcher();

		WaveStretcher(uint32 samplingRate, bool useQuick = false);

		CWaveStretcher* _detail() const;
	};

	/// <summary>
	/// 
	/// </summary>
	namespace Waving
	{
		/// <summary>
		/// テンポを変更した Wave を返します。
		/// </summary>
		/// <param name="wave">
		/// 元の Wave
		/// </param>
		/// <param name="tempo">
		/// 1.0 を基準としたテンポ。小さいと遅く、大きいと速くなる　[0.1, 10.0]
		/// </param>
		/// <param name="quickSeek">
		/// 品質を少し落として処理を高速化するかのフラグ
		/// </param>
		/// <returns>
		/// テンポを変更した Wave
		/// </returns>
		void ChangeTempo(const Wave& src, Wave& dst, double tempo, bool quickSeek = false);

		size_t ChangeTempo(WaveStretcher& stretcher, const WaveSample* src, size_t srcLength, WaveSample* dst, size_t dstLength, double tempo);

		/// <summary>
		/// ピッチを変更した Wave を返します。
		/// </summary>
		/// <param name="wave">
		/// 元の Wave
		/// </param>
		/// <param name="pitchSemitones">
		/// ピッチの変更 (半音上げる場合は 1.0, 1 オクターブ上げる場合は 12.0, 半音下げる場合は -1.0)
		/// </param>
		/// <param name="quickSeek">
		/// 品質を少し落として処理を高速化するかのフラグ
		/// </param>
		/// <returns>
		/// ピッチを変更した Wave
		/// </returns>
		void ChangePitchSemitones(const Wave& src, Wave& dst, double pitchSemitones, bool quickSeek = false);

		size_t ChangePitchSemitones(WaveStretcher& stretcher, const WaveSample* src, size_t srcLength, WaveSample* dst, size_t dstLength, double pitchSemitones);


		void ChangeTempoAndPitchSemitones(const Wave& src, Wave& dst, double tempo, double pitchSemitones, bool quickSeek = false);

		size_t ChangeTempoAndPitchSemitones(WaveStretcher& stretcher, const WaveSample* src, size_t srcLength, WaveSample* dst, size_t dstLength, double tempo, double pitchSemitones);


		/// <summary>
		/// サンプリングレートを変更した Wave を返します。
		/// </summary>
		/// <param name="wave">
		/// 元の Wave
		/// </param>
		/// <param name="samplingRate">
		/// サンプリングレート
		/// </param>
		/// <param name="quickSeek">
		/// 品質を少し落として処理を高速化するかのフラグ
		/// </param>
		/// <returns>
		/// サンプリングレートを変更した Wave
		/// </returns>
		void ChangeSamplingRate(const Wave& src, Wave& dst, uint32 samplingRate, bool quickSeek = false);

		WaveSample DoubleToSample(double mono);

		WaveSample DoubleToSample(double left, double right);

		void SwapChannel(const Wave& src, Wave& dst);

		void Reverse(const Wave& wave);

		inline double SinWave440Hz(double t)
		{
			return Math::Sin(t*440.0*Math::TwoPi);
		}

		Wave Synthesize(double lengthSec, std::function<double(double)> function = SinWave440Hz, uint32 rate = Wave::DefaultSamplingRate);
	}
}
